from enum import IntEnum
from typing import List, Optional

import numpy as np
import torch
from torch import Tensor
from torch.nn.utils.rnn import pad_sequence


class PadDirection(IntEnum):
    BEFORE = 0
    AFTER = 1


def convert_with_dir(
    seq: np.ndarray, dtype: torch.dtype, time_dim: int, pad_dir: PadDirection
) -> Tensor:
    if pad_dir == PadDirection.BEFORE:
        return torch.as_tensor(seq, dtype=dtype).flip(time_dim)

    return torch.as_tensor(seq, dtype=dtype)


def pad_with_dir(
    seq_list: List[Tensor], time_dim: int, pad_dir: PadDirection, **kwargs
) -> Tensor:
    if pad_dir == PadDirection.BEFORE:
        return pad_sequence(
            seq_list,
            **kwargs,
        ).flip(time_dim)

    return pad_sequence(
        seq_list,
        **kwargs,
    )


def pad_sequences(
    seq_list: List[np.ndarray],
    dtype: torch.dtype,
    time_dim: int,
    pad_dir: PadDirection,
    **kwargs,
) -> Tensor:
    return pad_with_dir(
        [convert_with_dir(seq, dtype, time_dim, pad_dir) for seq in seq_list],
        time_dim,
        pad_dir,
        **kwargs,
    )


def mask_up_to(lens: Tensor, delta: int = 0, max_len: Optional[int] = None) -> Tensor:
    """Exclusive.

    Args:
        lens (Tensor): _description_
        delta (int, optional): _description_. Defaults to 0.

    Returns:
        Tensor: _description_
    """
    if max_len is None:
        max_len = lens.max()

    arange_t: Tensor = torch.arange(
        max_len, dtype=lens.dtype, device=lens.device
    ).expand(*lens.shape, -1)

    return arange_t < (lens.unsqueeze(-1) + delta)


def vrange(starts: np.ndarray, stops: np.ndarray) -> np.ndarray:
    """Create concatenated ranges of integers for multiple start/stop

    Parameters:
        starts (1-D array): starts for each range
        stops (1-D array): stops for each range (same shape as starts)

    Returns:
        numpy.ndarray: concatenated ranges

    For example:

        >>> starts = np.array([1, 3, 4, 6])
        >>> stops  = np.array([1, 5, 7, 6])
        >>> vrange(starts, stops)
        array([3, 4, 4, 5, 6])

    """
    lens = stops - starts
    return np.repeat(stops - lens.cumsum(), lens) + np.arange(lens.sum())


def angle_wrap(radians: np.ndarray) -> np.ndarray:
    """This function wraps angles to lie within [-pi, pi).

    Args:
        radians (np.ndarray): The input array of angles (in radians).

    Returns:
        np.ndarray: Wrapped angles that lie within [-pi, pi).
    """
    return (radians + np.pi) % (2 * np.pi) - np.pi


def rotation_matrix(angle: float) -> np.ndarray:
    """Creates a 2D rotation matrix.

    Args:
        angle (float): The angle to rotate points by.

    Returns:
        np.ndarray: The 2x2 rotation matrix.
    """
    return np.array(
        [
            [np.cos(angle), -np.sin(angle)],
            [np.sin(angle), np.cos(angle)],
        ]
    )


def transform_matrices(angles: Tensor, translations: Tensor) -> Tensor:
    """Creates a 3x3 transformation matrix for each angle and translation in the input.

    Args:
        angles (Tensor): The (N,)-shaped angles tensor to rotate points by.
        translations (Tensor): The (N,2)-shaped translations to shift points by.

    Returns:
        Tensor: The Nx3x3 transformation matrices.
    """
    cos_vals = torch.cos(angles)
    sin_vals = torch.sin(angles)
    last_rows = torch.tensor(
        [[0.0, 0.0, 1.0]], dtype=angles.dtype, device=angles.device
    ).expand((angles.shape[0], -1))
    return torch.stack(
        [
            torch.stack([cos_vals, -sin_vals, translations[:, 0]], dim=-1),
            torch.stack([sin_vals, cos_vals, translations[:, 1]], dim=-1),
            last_rows,
        ],
        dim=-2,
    )


def batch_nd_transform_points_np(points, Mat):
    ndim = Mat.shape[-1] - 1
    batch = list(range(Mat.ndim - 2)) + [Mat.ndim - 1] + [Mat.ndim - 2]
    Mat = np.transpose(Mat, batch)
    if points.ndim == Mat.ndim - 1:
        return (points[..., np.newaxis, :] @ Mat[..., :ndim, :ndim]).squeeze(-2) + Mat[
            ..., -1:, :ndim
        ].squeeze(-2)
    elif points.ndim == Mat.ndim:
        return (
            (points[..., np.newaxis, :] @ Mat[..., np.newaxis, :ndim, :ndim])
            + Mat[..., np.newaxis, -1:, :ndim]
        ).squeeze(-2)
    else:
        raise Exception("wrong shape")


def agent_aware_diff(values: np.ndarray, agent_ids: np.ndarray) -> np.ndarray:
    values_diff: np.ndarray = np.diff(
        values, axis=0, prepend=values[[0]] - (values[[1]] - values[[0]])
    )

    # The point of the border mask is to catch data like this:
    # index    agent_id     vx    vy
    #     0           1    7.3   9.1
    #     1           2    0.0   0.0
    #                  ...
    # As implemented, we're not touching the very last row (we don't care anyways
    # for agents detected only once at a single timestep) and we would currently only
    # return index 0 (since we chop off the top with the 1: in the slice below), but
    # we want to return 1 so that's why the + 1 at the end.
    border_mask: np.ndarray = np.nonzero(agent_ids[1:-1] != agent_ids[:-2])[0] + 1
    values_diff[border_mask] = values_diff[border_mask + 1]

    return values_diff


def batch_proj(x, line):
    # x:[batch,3], line:[batch,N,3]
    line_length = line.shape[-2]
    batch_dim = x.ndim - 1
    if isinstance(x, torch.Tensor):
        delta = line[..., 0:2] - torch.unsqueeze(x[..., 0:2], dim=-2).repeat(
            *([1] * batch_dim), line_length, 1
        )
        dis = torch.linalg.norm(delta, axis=-1)
        idx0 = torch.argmin(dis, dim=-1)
        idx = idx0.view(*line.shape[:-2], 1, 1).repeat(
            *([1] * (batch_dim + 1)), line.shape[-1]
        )
        line_min = torch.squeeze(torch.gather(line, -2, idx), dim=-2)
        dx = x[..., None, 0] - line[..., 0]
        dy = x[..., None, 1] - line[..., 1]
        delta_y = -dx * torch.sin(line_min[..., None, 2]) + dy * torch.cos(
            line_min[..., None, 2]
        )
        delta_x = dx * torch.cos(line_min[..., None, 2]) + dy * torch.sin(
            line_min[..., None, 2]
        )

        delta_psi = angle_wrap(x[..., 2] - line_min[..., 2])

        return (
            delta_x,
            delta_y,
            torch.unsqueeze(delta_psi, dim=-1),
        )
    elif isinstance(x, np.ndarray):
        delta = line[..., 0:2] - np.repeat(
            x[..., np.newaxis, 0:2], line_length, axis=-2
        )
        dis = np.linalg.norm(delta, axis=-1)
        idx0 = np.argmin(dis, axis=-1)
        idx = idx0.reshape(*line.shape[:-2], 1, 1).repeat(line.shape[-1], axis=-1)
        line_min = np.squeeze(np.take_along_axis(line, idx, axis=-2), axis=-2)
        dx = x[..., None, 0] - line[..., 0]
        dy = x[..., None, 1] - line[..., 1]
        delta_y = -dx * np.sin(line_min[..., None, 2]) + dy * np.cos(
            line_min[..., None, 2]
        )
        delta_x = dx * np.cos(line_min[..., None, 2]) + dy * np.sin(
            line_min[..., None, 2]
        )
        delta_psi = angle_wrap(x[..., 2] - line_min[..., 2])
        return (
            delta_x,
            delta_y,
            np.expand_dims(delta_psi, axis=-1),
        )



from typing import Optional, Type, Union

from trajdata.caching import EnvCache, SceneCache
from trajdata.data_structures import Scene, SceneMetadata
from trajdata.dataset_specific import RawDataset
from trajdata.utils import scene_utils


def get_agent_data(
    scene_info: SceneMetadata,
    raw_dataset: RawDataset,
    env_cache: EnvCache,
    rebuild_cache: bool,
    cache_class: Type[SceneCache],
    desired_dt: Optional[float] = None,
) -> Scene:
    if not rebuild_cache and env_cache.scene_is_cached(
        scene_info.env_name, scene_info.name, scene_info.dt
    ):
        scene: Scene = env_cache.load_scene(
            scene_info.env_name, scene_info.name, scene_info.dt
        )

        # If the original data is already cached...
        if scene_utils.enforce_desired_dt(scene_info, desired_dt, dry_run=True):
            # If the original data is already cached,
            # but this scene's dt doesn't match what we desire:
            # First, interpolate and save the data.
            # Then, return the interpolated scene.

            # Interpolating the scene metadata and caching it.
            scene_utils.enforce_desired_dt(scene, desired_dt)
            env_cache.save_scene(scene)

            # Interpolating the agent data and caching it.
            # The core point of doing this here rather than in Line 45 and below
            # is that we do not need to access the raw dataset object, we can
            # leverage the already cached data.
            scene_cache: SceneCache = cache_class(env_cache.path, scene)
            scene_cache.write_cache_to_disk()

        # Once this scene's dt matches what we desire: Return it.
        return scene

    # Obtaining and caching the original scene data.
    scene: Scene = raw_dataset.get_scene(scene_info)
    agent_list, agent_presence = raw_dataset.get_agent_info(
        scene, env_cache.path, cache_class
    )

    scene.update_agent_info(agent_list, agent_presence)
    env_cache.save_scene(scene)

    if scene_utils.enforce_desired_dt(scene, desired_dt, dry_run=True):
        # In case the user specified a desired_dt that's different from the scene's
        # native dt, we will perform the interpolation here and cache the result for
        # later reuse.

        # Interpolating the scene metadata and caching it.
        scene_utils.enforce_desired_dt(scene, desired_dt)
        env_cache.save_scene(scene)

        # Interpolating the agent data and caching it.
        scene_cache: SceneCache = cache_class(env_cache.path, scene)
        scene_cache.write_cache_to_disk()

    return scene



from math import ceil
from typing import Any, Dict, Final, List, Optional, Tuple

import cv2
import numpy as np
from tqdm import tqdm

from trajdata.proto.vectorized_map_pb2 import (
    MapElement,
    Polyline,
    RoadLane,
    VectorizedMap,
)

# Sub-pixel drawing precision constants.
# See https://github.com/woven-planet/l5kit/blob/master/l5kit/l5kit/rasterization/semantic_rasterizer.py#L16
CV2_SUB_VALUES = {"shift": 9, "lineType": cv2.LINE_AA}
CV2_SHIFT_VALUE = 2 ** CV2_SUB_VALUES["shift"]

MM_PER_M: Final[float] = 1000


def cv2_subpixel(coords: np.ndarray) -> np.ndarray:
    """
    Cast coordinates to numpy.int but keep fractional part by previously multiplying by 2**CV2_SHIFT
    cv2 calls will use shift to restore original values with higher precision

    Args:
        coords (np.ndarray): XY coords as float

    Returns:
        np.ndarray: XY coords as int for cv2 shift draw
    """
    return (coords * CV2_SHIFT_VALUE).astype(np.int)


def decompress_values(data: np.ndarray) -> np.ndarray:
    # From https://github.com/woven-planet/l5kit/blob/master/l5kit/l5kit/data/proto/road_network.proto#L446
    # The delta for the first point is just its coordinates tuple, i.e. it is a "delta" from
    # the origin. For subsequent points, this field stores the difference between the point's
    # coordinates and the previous point's coordinates. This is for representation efficiency.
    return np.cumsum(data, axis=0, dtype=np.float) / MM_PER_M


def compress_values(data: np.ndarray) -> np.ndarray:
    return (np.diff(data, axis=0, prepend=0.0) * MM_PER_M).astype(np.int32)


def populate_lane_polylines(
    new_lane: RoadLane,
    midlane_pts: np.ndarray,
    left_pts: np.ndarray,
    right_pts: np.ndarray,
) -> None:
    """Fill a Lane object's polyline attributes.
    All points should be in world coordinates.

    Args:
        new_lane (Lane): _description_
        midlane_pts (np.ndarray): _description_
        left_pts (np.ndarray): _description_
        right_pts (np.ndarray): _description_
    """
    compressed_mid_pts: np.ndarray = compress_values(midlane_pts)
    compressed_left_pts: np.ndarray = compress_values(left_pts)
    compressed_right_pts: np.ndarray = compress_values(right_pts)

    new_lane.center.dx_mm.extend(compressed_mid_pts[:, 0].tolist())
    new_lane.center.dy_mm.extend(compressed_mid_pts[:, 1].tolist())

    new_lane.left_boundary.dx_mm.extend(compressed_left_pts[:, 0].tolist())
    new_lane.left_boundary.dy_mm.extend(compressed_left_pts[:, 1].tolist())

    new_lane.right_boundary.dx_mm.extend(compressed_right_pts[:, 0].tolist())
    new_lane.right_boundary.dy_mm.extend(compressed_right_pts[:, 1].tolist())

    if compressed_mid_pts.shape[-1] == 3:
        new_lane.center.dz_mm.extend(compressed_mid_pts[:, 2].tolist())
        new_lane.left_boundary.dz_mm.extend(compressed_left_pts[:, 2].tolist())
        new_lane.right_boundary.dz_mm.extend(compressed_right_pts[:, 2].tolist())


def populate_polygon(
    polygon: Polyline,
    polygon_pts: np.ndarray,
) -> None:
    """Fill a Crosswalk object's polygon attribute.
    All points should be in world coordinates.

    Args:
        new_crosswalk (Lane): _description_
        polygon_pts (np.ndarray): _description_
    """

    compressed_pts: np.ndarray = compress_values(polygon_pts)

    polygon.dx_mm.extend(compressed_pts[:, 0].tolist())
    polygon.dy_mm.extend(compressed_pts[:, 1].tolist())

    if compressed_pts.shape[-1] == 3:
        polygon.dz_mm.extend(compressed_pts[:, 2].tolist())


def proto_to_np(polyline: Polyline) -> np.ndarray:
    dx: np.ndarray = np.asarray(polyline.dx_mm)
    dy: np.ndarray = np.asarray(polyline.dy_mm)

    if len(polyline.dz_mm) > 0:
        dz: np.ndarray = np.asarray(polyline.dz_mm)
        pts: np.ndarray = np.stack([dx, dy, dz], axis=1)
    else:
        pts: np.ndarray = np.stack([dx, dy], axis=1)

    return decompress_values(pts)


def transform_points(points: np.ndarray, transf_mat: np.ndarray):
    n_dim = points.shape[-1]
    return points @ transf_mat[:n_dim, :n_dim] + transf_mat[:n_dim, -1]


def interpolate(pts: np.ndarray, num_pts: int) -> np.ndarray:
    """
    Interpolate points based on cumulative distances from the first one. In particular,
    interpolate using a variable step such that we always get step values.

    Args:
        xyz (np.ndarray): XYZ coords.
        num_pts (int): How many points to interpolate to.

    Returns:
        np.ndarray: The new interpolated coordinates.
    """
    cum_dist = np.cumsum(np.linalg.norm(np.diff(pts, axis=0), axis=-1))
    cum_dist = np.insert(cum_dist, 0, 0)

    assert num_pts > 1, f"num_pts must be at least 2, but got {num_pts}"
    steps = np.linspace(cum_dist[0], cum_dist[-1], num_pts)

    xyz_inter = np.empty((len(steps), pts.shape[-1]), dtype=pts.dtype)
    xyz_inter[:, 0] = np.interp(steps, xp=cum_dist, fp=pts[:, 0])
    xyz_inter[:, 1] = np.interp(steps, xp=cum_dist, fp=pts[:, 1])
    if pts.shape[-1] == 3:
        xyz_inter[:, 2] = np.interp(steps, xp=cum_dist, fp=pts[:, 2])

    return xyz_inter


def rasterize_map(
    vec_map: VectorizedMap, resolution: float, **pbar_kwargs
) -> np.ndarray:
    """Renders the semantic map at the given resolution.

    Args:
        vec_map (VectorizedMap): _description_
        resolution (float): The rasterized image's resolution in pixels per meter.

    Returns:
        np.ndarray: The rasterized RGB image.
    """
    world_center_m: Tuple[float, float] = (
        (vec_map.max_pt.x + vec_map.min_pt.x) / 2,
        (vec_map.max_pt.y + vec_map.min_pt.y) / 2,
    )

    raster_size_x: int = ceil((vec_map.max_pt.x - vec_map.min_pt.x) * resolution)
    raster_size_y: int = ceil((vec_map.max_pt.y - vec_map.min_pt.y) * resolution)

    raster_from_local: np.ndarray = np.array(
        [
            [resolution, 0, raster_size_x / 2],
            [0, resolution, raster_size_y / 2],
            [0, 0, 1],
        ]
    )

    # Compute pose from its position and rotation
    pose_from_world: np.ndarray = np.array(
        [
            [1, 0, -world_center_m[0]],
            [0, 1, -world_center_m[1]],
            [0, 0, 1],
        ]
    )

    raster_from_world: np.ndarray = raster_from_local @ pose_from_world

    lane_area_img: np.ndarray = np.zeros(
        shape=(raster_size_y, raster_size_x, 3), dtype=np.uint8
    )
    lane_line_img: np.ndarray = np.zeros(
        shape=(raster_size_y, raster_size_x, 3), dtype=np.uint8
    )
    ped_area_img: np.ndarray = np.zeros(
        shape=(raster_size_y, raster_size_x, 3), dtype=np.uint8
    )

    map_elem: MapElement
    for map_elem in tqdm(
        vec_map.elements,
        desc=f"Rasterizing Map at {resolution:.2f} px/m",
        **pbar_kwargs,
    ):
        if map_elem.HasField("road_lane"):
            left_pts: np.ndarray = proto_to_np(map_elem.road_lane.left_boundary)
            right_pts: np.ndarray = proto_to_np(map_elem.road_lane.right_boundary)

            lane_area: np.ndarray = cv2_subpixel(
                transform_points(
                    np.concatenate([left_pts[:, :2], right_pts[::-1, :2]], axis=0),
                    raster_from_world,
                )
            )

            # Need to for-loop because doing it all at once can make holes.
            cv2.fillPoly(
                img=lane_area_img,
                pts=[lane_area],
                color=(255, 0, 0),
                **CV2_SUB_VALUES,
            )

            # Drawing lane lines.
            cv2.polylines(
                img=lane_line_img,
                pts=lane_area.reshape((2, -1, 2)),
                isClosed=False,
                color=(0, 255, 0),
                **CV2_SUB_VALUES,
            )

        elif map_elem.HasField("road_area"):
            xyz_pts: np.ndarray = proto_to_np(map_elem.road_area.exterior_polygon)
            road_area: np.ndarray = cv2_subpixel(
                transform_points(xyz_pts[:, :2], raster_from_world)
            )

            # Drawing general road areas.
            cv2.fillPoly(
                img=lane_area_img,
                pts=[road_area],
                color=(255, 0, 0),
                **CV2_SUB_VALUES,
            )

            for interior_hole in map_elem.road_area.interior_holes:
                xyz_pts: np.ndarray = proto_to_np(interior_hole)
                road_area: np.ndarray = cv2_subpixel(
                    transform_points(xyz_pts[:, :2], raster_from_world)
                )

                # Removing holes.
                cv2.fillPoly(
                    img=lane_area_img,
                    pts=[road_area],
                    color=(0, 0, 0),
                    **CV2_SUB_VALUES,
                )

        elif map_elem.HasField("ped_crosswalk"):
            xyz_pts: np.ndarray = proto_to_np(map_elem.ped_crosswalk.polygon)
            crosswalk_area: np.ndarray = cv2_subpixel(
                transform_points(xyz_pts[:, :2], raster_from_world)
            )

            # Drawing crosswalks.
            cv2.fillPoly(
                img=ped_area_img,
                pts=[crosswalk_area],
                color=(0, 0, 255),
                **CV2_SUB_VALUES,
            )

        elif map_elem.HasField("ped_walkway"):
            xyz_pts: np.ndarray = proto_to_np(map_elem.ped_walkway.polygon)
            walkway_area: np.ndarray = cv2_subpixel(
                transform_points(xyz_pts[:, :2], raster_from_world)
            )

            # Drawing walkways.
            cv2.fillPoly(
                img=ped_area_img,
                pts=[walkway_area],
                color=(0, 0, 255),
                **CV2_SUB_VALUES,
            )

    map_img: np.ndarray = (lane_area_img + lane_line_img + ped_area_img).astype(
        np.float32
    ) / 255
    return map_img.transpose(2, 0, 1), raster_from_world



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vectorized_map.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x14vectorized_map.proto\x12\x08trajdata"\x99\x01\n\rVectorizedMap\x12&\n\x08\x65lements\x18\x01 \x03(\x0b\x32\x14.trajdata.MapElement\x12$\n\x06max_pt\x18\x02 \x01(\x0b\x32\x0f.trajdata.PointH\x00\x88\x01\x01\x12$\n\x06min_pt\x18\x03 \x01(\x0b\x32\x0f.trajdata.PointH\x01\x88\x01\x01\x42\t\n\x07_max_ptB\t\n\x07_min_pt"\xd8\x01\n\nMapElement\x12\n\n\x02id\x18\x01 \x01(\x0c\x12\'\n\troad_lane\x18\x02 \x01(\x0b\x32\x12.trajdata.RoadLaneH\x00\x12\'\n\troad_area\x18\x03 \x01(\x0b\x32\x12.trajdata.RoadAreaH\x00\x12/\n\rped_crosswalk\x18\x04 \x01(\x0b\x32\x16.trajdata.PedCrosswalkH\x00\x12+\n\x0bped_walkway\x18\x05 \x01(\x0b\x32\x14.trajdata.PedWalkwayH\x00\x42\x0e\n\x0c\x65lement_data"I\n\x05Point\x12\x0e\n\x01x\x18\x01 \x01(\x01H\x00\x88\x01\x01\x12\x0e\n\x01y\x18\x02 \x01(\x01H\x01\x88\x01\x01\x12\x0e\n\x01z\x18\x03 \x01(\x01H\x02\x88\x01\x01\x42\x04\n\x02_xB\x04\n\x02_yB\x04\n\x02_z"7\n\x08Polyline\x12\r\n\x05\x64x_mm\x18\x01 \x03(\x11\x12\r\n\x05\x64y_mm\x18\x02 \x03(\x11\x12\r\n\x05\x64z_mm\x18\x03 \x03(\x11"\xe9\x01\n\x08RoadLane\x12"\n\x06\x63\x65nter\x18\x01 \x01(\x0b\x32\x12.trajdata.Polyline\x12)\n\rleft_boundary\x18\x02 \x01(\x0b\x32\x12.trajdata.Polyline\x12*\n\x0eright_boundary\x18\x03 \x01(\x0b\x32\x12.trajdata.Polyline\x12\x13\n\x0b\x65ntry_lanes\x18\x04 \x03(\x0c\x12\x12\n\nexit_lanes\x18\x05 \x03(\x0c\x12\x1b\n\x13\x61\x64jacent_lanes_left\x18\x06 \x03(\x0c\x12\x1c\n\x14\x61\x64jacent_lanes_right\x18\x07 \x03(\x0c"d\n\x08RoadArea\x12,\n\x10\x65xterior_polygon\x18\x01 \x01(\x0b\x32\x12.trajdata.Polyline\x12*\n\x0einterior_holes\x18\x02 \x03(\x0b\x32\x12.trajdata.Polyline"3\n\x0cPedCrosswalk\x12#\n\x07polygon\x18\x01 \x01(\x0b\x32\x12.trajdata.Polyline"1\n\nPedWalkway\x12#\n\x07polygon\x18\x01 \x01(\x0b\x32\x12.trajdata.Polylineb\x06proto3'
)


_VECTORIZEDMAP = DESCRIPTOR.message_types_by_name["VectorizedMap"]
_MAPELEMENT = DESCRIPTOR.message_types_by_name["MapElement"]
_POINT = DESCRIPTOR.message_types_by_name["Point"]
_POLYLINE = DESCRIPTOR.message_types_by_name["Polyline"]
_ROADLANE = DESCRIPTOR.message_types_by_name["RoadLane"]
_ROADAREA = DESCRIPTOR.message_types_by_name["RoadArea"]
_PEDCROSSWALK = DESCRIPTOR.message_types_by_name["PedCrosswalk"]
_PEDWALKWAY = DESCRIPTOR.message_types_by_name["PedWalkway"]
VectorizedMap = _reflection.GeneratedProtocolMessageType(
    "VectorizedMap",
    (_message.Message,),
    {
        "DESCRIPTOR": _VECTORIZEDMAP,
        "__module__": "vectorized_map_pb2"
        # @@protoc_insertion_point(class_scope:trajdata.VectorizedMap)
    },
)
_sym_db.RegisterMessage(VectorizedMap)

MapElement = _reflection.GeneratedProtocolMessageType(
    "MapElement",
    (_message.Message,),
    {
        "DESCRIPTOR": _MAPELEMENT,
        "__module__": "vectorized_map_pb2"
        # @@protoc_insertion_point(class_scope:trajdata.MapElement)
    },
)
_sym_db.RegisterMessage(MapElement)

Point = _reflection.GeneratedProtocolMessageType(
    "Point",
    (_message.Message,),
    {
        "DESCRIPTOR": _POINT,
        "__module__": "vectorized_map_pb2"
        # @@protoc_insertion_point(class_scope:trajdata.Point)
    },
)
_sym_db.RegisterMessage(Point)

Polyline = _reflection.GeneratedProtocolMessageType(
    "Polyline",
    (_message.Message,),
    {
        "DESCRIPTOR": _POLYLINE,
        "__module__": "vectorized_map_pb2"
        # @@protoc_insertion_point(class_scope:trajdata.Polyline)
    },
)
_sym_db.RegisterMessage(Polyline)

RoadLane = _reflection.GeneratedProtocolMessageType(
    "RoadLane",
    (_message.Message,),
    {
        "DESCRIPTOR": _ROADLANE,
        "__module__": "vectorized_map_pb2"
        # @@protoc_insertion_point(class_scope:trajdata.RoadLane)
    },
)
_sym_db.RegisterMessage(RoadLane)

RoadArea = _reflection.GeneratedProtocolMessageType(
    "RoadArea",
    (_message.Message,),
    {
        "DESCRIPTOR": _ROADAREA,
        "__module__": "vectorized_map_pb2"
        # @@protoc_insertion_point(class_scope:trajdata.RoadArea)
    },
)
_sym_db.RegisterMessage(RoadArea)

PedCrosswalk = _reflection.GeneratedProtocolMessageType(
    "PedCrosswalk",
    (_message.Message,),
    {
        "DESCRIPTOR": _PEDCROSSWALK,
        "__module__": "vectorized_map_pb2"
        # @@protoc_insertion_point(class_scope:trajdata.PedCrosswalk)
    },
)
_sym_db.RegisterMessage(PedCrosswalk)

PedWalkway = _reflection.GeneratedProtocolMessageType(
    "PedWalkway",
    (_message.Message,),
    {
        "DESCRIPTOR": _PEDWALKWAY,
        "__module__": "vectorized_map_pb2"
        # @@protoc_insertion_point(class_scope:trajdata.PedWalkway)
    },
)
_sym_db.RegisterMessage(PedWalkway)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _VECTORIZEDMAP._serialized_start = 35
    _VECTORIZEDMAP._serialized_end = 188
    _MAPELEMENT._serialized_start = 191
    _MAPELEMENT._serialized_end = 407
    _POINT._serialized_start = 409
    _POINT._serialized_end = 482
    _POLYLINE._serialized_start = 484
    _POLYLINE._serialized_end = 539
    _ROADLANE._serialized_start = 542
    _ROADLANE._serialized_end = 775
    _ROADAREA._serialized_start = 777
    _ROADAREA._serialized_end = 877
    _PEDCROSSWALK._serialized_start = 879
    _PEDCROSSWALK._serialized_end = 930
    _PEDWALKWAY._serialized_start = 932
    _PEDWALKWAY._serialized_end = 981
# @@protoc_insertion_point(module_scope)



from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

from trajdata import filtering
from trajdata.augmentation import BatchAugmentation
from trajdata.caching.df_cache import DataFrameCache
from trajdata.data_structures.agent import AgentMetadata, FixedExtent, VariableExtent
from trajdata.data_structures.batch import AgentBatch
from trajdata.data_structures.batch_element import AgentBatchElement
from trajdata.data_structures.collation import agent_collate_fn
from trajdata.data_structures.scene import SceneTimeAgent
from trajdata.data_structures.scene_metadata import Scene
from trajdata.dataset import UnifiedDataset
from trajdata.simulation.sim_cache import SimulationCache
from trajdata.simulation.sim_df_cache import SimulationDataFrameCache
from trajdata.simulation.sim_metrics import SimMetric
from trajdata.simulation.sim_stats import SimStatistic


class SimulationScene:
    def __init__(
        self,
        env_name: str,
        scene_name: str,
        scene: Scene,
        dataset: UnifiedDataset,
        init_timestep: int = 0,
        freeze_agents: bool = True,
        return_dict: bool = False,
    ) -> None:
        if not freeze_agents:
            raise NotImplementedError(
                (
                    "Agents that change over time (i.e., following the original dataset) "
                    "are not handled yet internally. Please set freeze_agents=True."
                )
            )

        self.env_name: str = env_name
        self.scene_name: str = scene_name
        self.scene: Scene = deepcopy(scene)
        self.dataset: UnifiedDataset = dataset
        self.init_scene_ts: int = init_timestep
        self.freeze_agents: bool = freeze_agents
        self.return_dict: bool = return_dict
        self.scene_ts: int = self.init_scene_ts

        agents_present: List[AgentMetadata] = self.scene.agent_presence[self.scene_ts]
        self.agents: List[AgentMetadata] = filtering.agent_types(
            agents_present, self.dataset.no_types, self.dataset.only_types
        )

        if len(self.agents) == 0:
            raise ValueError(
                (
                    f"Initial timestep {self.scene_ts} contains no agents after filtering. "
                    "Please choose another initial timestep."
                )
            )

        if self.freeze_agents:
            self.scene.agent_presence = self.scene.agent_presence[
                : self.init_scene_ts + 1
            ]
            self.scene.agents = self.agents

        # Note this order of operations is important, we first instantiate
        # the cache with the copied scene_info + modified agents list.
        # Then, we change the env_name and etc later during finalization
        # (if we did it earlier then the cache would go looking inside
        # the sim folder for scene data rather than the original scene
        # data location).
        if self.dataset.cache_class == DataFrameCache:
            self.cache: SimulationCache = SimulationDataFrameCache(
                dataset.cache_path,
                self.scene,
                init_timestep,
                dataset.augmentations,
            )

        self.batch_augments: Optional[List[BatchAugmentation]] = None
        if dataset.augmentations:
            self.batch_augments = [
                batch_aug
                for batch_aug in dataset.augmentations
                if isinstance(batch_aug, BatchAugmentation)
            ]

    def reset(self) -> Union[AgentBatch, Dict[str, Any]]:
        self.scene_ts: int = self.init_scene_ts
        return self.get_obs()

    def step(
        self,
        new_xyh_dict: Dict[str, np.ndarray],
        return_obs=True,
    ) -> Union[AgentBatch, Dict[str, Any]]:
        self.scene_ts += 1

        self.cache.append_state(new_xyh_dict)

        if not self.freeze_agents:
            agents_present: List[AgentMetadata] = self.scene.agent_presence[
                self.scene_ts
            ]
            self.agents: List[AgentMetadata] = filtering.agent_types(
                agents_present, self.dataset.no_types, self.dataset.only_types
            )

            self.scene.agent_presence[self.scene_ts] = self.agents
        else:
            self.scene.agent_presence.append(self.agents)

        if return_obs:
            return self.get_obs()

    def get_obs(
        self, collate: bool = True, get_map: bool = True
    ) -> Union[AgentBatch, Dict[str, Any]]:
        agent_data_list: List[AgentBatchElement] = list()
        for agent in self.agents:
            scene_time_agent = SceneTimeAgent(
                self.scene, self.scene_ts, self.agents, agent, self.cache
            )
            agent_data_list.append(
                AgentBatchElement(
                    self.cache,
                    -1,  # Not used
                    scene_time_agent,
                    history_sec=self.dataset.history_sec,
                    future_sec=self.dataset.future_sec,
                    agent_interaction_distances=self.dataset.agent_interaction_distances,
                    incl_robot_future=False,
                    incl_map=get_map and self.dataset.incl_map,
                    map_params=self.dataset.map_params,
                    standardize_data=self.dataset.standardize_data,
                    standardize_derivatives=self.dataset.standardize_derivatives,
                    max_neighbor_num=self.dataset.max_neighbor_num,
                )
            )

            # Need to reset transformations for each agent since each
            # AgentBatchElement transforms (standardizes) the cache.
            self.cache.reset_transforms()

        if collate:
            return agent_collate_fn(
                agent_data_list,
                return_dict=self.return_dict,
                pad_format="outside",
                batch_augments=self.batch_augments,
            )
        else:
            return agent_data_list

    def get_metrics(self, metrics: List[SimMetric]) -> Dict[str, Dict[str, float]]:
        return self.cache.calculate_metrics(
            metrics, ts_range=(self.init_scene_ts + 1, self.scene_ts)
        )

    def get_stats(
        self, stats: List[SimStatistic]
    ) -> Dict[str, Dict[str, Tuple[np.ndarray, np.ndarray]]]:
        return self.cache.calculate_stats(
            stats, ts_range=(self.init_scene_ts + 1, self.scene_ts)
        )

    def finalize(self) -> None:
        # We only change the agent's last timestep here because we use it
        # earlier to check if the agent has any future data from the original
        # dataset.
        for agent in self.agents:
            agent.last_timestep = self.scene_ts

        self.scene.length_timesteps = self.scene_ts + 1

        self.scene.agent_presence = self.scene.agent_presence[: self.scene_ts + 1]

        self.scene.env_metadata.name = self.env_name
        self.scene.env_name = self.env_name
        self.scene.name = self.scene_name

    def save(self) -> None:
        self.dataset.env_cache.save_scene(self.scene)
        self.cache.save_sim_scene(self.scene)

    def add_new_agents(self, agent_data: List[Tuple]):
        existing_agent_names = [agent.name for agent in self.agents]
        agent_data = [
            agent for agent in agent_data if agent[0] not in existing_agent_names
        ]
        if len(agent_data) > 0:
            self.cache.add_agents(agent_data)
            for data in agent_data:
                name, state, ts0, agent_type, extent = data
                metadata = AgentMetadata(
                    name=name,
                    agent_type=agent_type,
                    first_timestep=ts0,
                    last_timestep=ts0 + state.shape[0] - 1,
                    extent=FixedExtent(
                        length=extent[0], width=extent[1], height=extent[2]
                    ),
                )
                self.agents.append(metadata)



from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple

import numpy as np

from trajdata.augmentation.augmentation import Augmentation
from trajdata.data_structures.agent import AgentMetadata
from trajdata.data_structures.scene_metadata import Scene
from trajdata.maps import RasterizedMap, RasterizedMapMetadata
from trajdata.proto.vectorized_map_pb2 import VectorizedMap


class SceneCache:
    def __init__(
        self,
        cache_path: Path,
        scene: Scene,
        scene_ts: Optional[int] = 0,
        augmentations: Optional[List[Augmentation]] = None,
    ) -> None:
        """
        Creates and prepares the cache for online data loading.
        """
        self.path = cache_path
        self.scene = scene
        self.dt = scene.dt
        self.scene_ts = scene_ts
        self.augmentations = augmentations

        # Ensuring the scene cache folder exists
        self.scene_dir: Path = self.path / self.scene.env_name / self.scene.name
        self.scene_dir.mkdir(parents=True, exist_ok=True)

    def write_cache_to_disk(self) -> None:
        """Saves agent data to disk for fast loading later (just like save_agent_data),
        but using the class attributes for the sources of data and file paths.
        """
        raise NotImplementedError()

    # AGENT STATE DATA
    @staticmethod
    def save_agent_data(
        agent_data: Any,
        cache_path: Path,
        scene: Scene,
    ) -> None:
        """Saves agent data to disk for fast loading later."""
        raise NotImplementedError()

    def get_value(self, agent_id: str, scene_ts: int, attribute: str) -> float:
        """
        Get a single attribute value for an agent at a timestep.
        """
        raise NotImplementedError()

    def get_state(self, agent_id: str, scene_ts: int) -> np.ndarray:
        """
        Get an agent's state at a specific timestep.
        """
        raise NotImplementedError()

    def get_states(self, agent_ids: List[str], scene_ts: int) -> np.ndarray:
        """
        Get multiple agents' states at a specific timestep.
        """
        raise NotImplementedError()

    def transform_data(self, **kwargs) -> None:
        """
        Transform the data before accessing it later, e.g., to make the mean zero or rotate the scene around an agent.
        This can either be done in this function call or just stored for later lazy application.
        """
        raise NotImplementedError()

    def reset_transforms(self) -> None:
        """
        Transform the data back to its original coordinate system.
        This can either be done in this function call or just stored for later lazy application.
        """
        raise NotImplementedError()

    def interpolate_data(self, desired_dt: float, method: str = "linear") -> None:
        """Increase the sampling frequency of the data by interpolation.

        Args:
            desired_dt (float): The desired spacing between timesteps.
            method (str, optional): The type of interpolation to use, currently only "linear" is implemented. Defaults to "linear".
        """
        raise NotImplementedError()

    def get_agent_history(
        self,
        agent_info: AgentMetadata,
        scene_ts: int,
        history_sec: Tuple[Optional[float], Optional[float]],
    ) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError()

    def get_agent_future(
        self,
        agent_info: AgentMetadata,
        scene_ts: int,
        future_sec: Tuple[Optional[float], Optional[float]],
    ) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError()

    def get_agents_history(
        self,
        scene_ts: int,
        agents: List[AgentMetadata],
        history_sec: Tuple[Optional[float], Optional[float]],
    ) -> Tuple[List[np.ndarray], List[np.ndarray], np.ndarray]:
        raise NotImplementedError()

    def get_agents_future(
        self,
        scene_ts: int,
        agents: List[AgentMetadata],
        future_sec: Tuple[Optional[float], Optional[float]],
    ) -> Tuple[List[np.ndarray], List[np.ndarray], np.ndarray]:
        raise NotImplementedError()

    # MAPS
    @staticmethod
    def are_maps_cached(cache_path: Path, env_name: str) -> bool:
        raise NotImplementedError()

    @staticmethod
    def is_map_cached(
        cache_path: Path, env_name: str, map_name: str, resolution: float
    ) -> bool:
        raise NotImplementedError()

    @staticmethod
    def cache_map(
        cache_path: Path, vec_map: VectorizedMap, map_obj: RasterizedMap, env_name: str
    ) -> None:
        raise NotImplementedError()

    @staticmethod
    def cache_map_layers(
        cache_path: Path,
        vec_map: VectorizedMap,
        map_info: RasterizedMapMetadata,
        layer_fn: Callable[[str], np.ndarray],
        env_name: str,
    ) -> None:
        raise NotImplementedError()

    def load_map_patch(
        self,
        world_x: float,
        world_y: float,
        desired_patch_size: int,
        resolution: float,
        offset_xy: Tuple[float, float],
        agent_heading: float,
        return_rgb: bool,
        rot_pad_factor: float = 1.0,
        no_map_val: float = 0.0,
    ) -> Tuple[np.ndarray, np.ndarray, bool]:
        raise NotImplementedError()



from typing import List, Optional, Tuple

import numpy as np
import torch
from torch import Tensor


class MapMetadata:
    def __init__(
        self,
        name: str,
        shape: Tuple[int, int],
        layers: List[str],
        layer_rgb_groups: Tuple[List[int], List[int], List[int]],
        resolution: float,  # px/m
        map_from_world: np.ndarray,  # Transformation from world coordinates [m] to map coordinates [px]
    ) -> None:
        self.name: str = name
        self.shape: Tuple[int, int] = shape
        self.layers: List[str] = layers
        self.layer_rgb_groups: Tuple[List[int], List[int], List[int]] = layer_rgb_groups
        self.resolution: float = resolution
        self.map_from_world: np.ndarray = map_from_world


class Map:
    def __init__(
        self,
        metadata: MapMetadata,
        data: np.ndarray,
    ) -> None:
        assert data.shape == metadata.shape
        self.metadata: MapMetadata = metadata
        self.data: np.ndarray = data

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.data.shape

    @staticmethod
    def to_img(
        map_arr: Tensor,
        idx_groups: Optional[Tuple[List[int], List[int], List[int]]] = None,
    ) -> Tensor:
        if idx_groups is None:
            return map_arr.permute(1, 2, 0).numpy()

        return torch.stack(
            [
                torch.amax(map_arr[idx_groups[0]], dim=0),
                torch.amax(map_arr[idx_groups[1]], dim=0),
                torch.amax(map_arr[idx_groups[2]], dim=0),
            ],
            dim=-1,
        ).numpy()



from .agent import Agent, AgentMetadata, AgentType, FixedExtent, VariableExtent
from .batch import AgentBatch, SceneBatch
from .batch_element import AgentBatchElement, SceneBatchElement
from .collation import agent_collate_fn, scene_collate_fn
from .data_index import AgentDataIndex, DataIndex, SceneDataIndex
from .environment import EnvMetadata
from .scene import SceneTime, SceneTimeAgent
from .scene_metadata import Scene, SceneMetadata
from .scene_tag import SceneTag



import itertools
from pathlib import Path
from typing import Dict, List, Tuple

from trajdata.data_structures.scene_tag import SceneTag


class EnvMetadata:
    def __init__(
        self,
        name: str,
        data_dir: str,
        dt: float,
        parts: List[Tuple[str]],
        scene_split_map: Dict[str, str],
    ) -> None:
        self.name = name
        self.data_dir = Path(data_dir).expanduser().resolve()
        self.dt = dt
        self.parts = parts
        self.scene_tags: List[SceneTag] = [
            SceneTag(tag_tuple)
            # Cartesian product of the given list of tuples
            for tag_tuple in itertools.product(*([(name,)] + parts))
        ]
        self.scene_split_map = scene_split_map



from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
import torch
from torch import Tensor

from trajdata.utils import arr_utils


class SimStatistic:
    def __init__(self, name: str) -> None:
        self.name = name

    def __call__(self, scene_df: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError()


class VelocityHistogram(SimStatistic):
    def __init__(self, bins: List[int]) -> None:
        super().__init__("vel_hist")
        self.bins = bins

    def __call__(self, scene_df: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        velocities: np.ndarray = np.linalg.norm(scene_df[["vx", "vy"]], axis=1)

        return np.histogram(velocities, bins=self.bins)


class LongitudinalAccHistogram(SimStatistic):
    def __init__(self, bins: List[int]) -> None:
        super().__init__("lon_acc_hist")
        self.bins = bins

    def __call__(self, scene_df: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        accels: np.ndarray = np.linalg.norm(scene_df[["ax", "ay"]], axis=1)
        lon_accels: np.ndarray = accels * np.cos(scene_df["heading"])

        return np.histogram(lon_accels, bins=self.bins)


class LateralAccHistogram(SimStatistic):
    def __init__(self, bins: List[int]) -> None:
        super().__init__("lat_acc_hist")
        self.bins = bins

    def __call__(self, scene_df: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        accels: np.ndarray = np.linalg.norm(scene_df[["ax", "ay"]], axis=1)
        lat_accels: np.ndarray = accels * np.sin(scene_df["heading"])

        return np.histogram(lat_accels, bins=self.bins)


class JerkHistogram(SimStatistic):
    def __init__(self, bins: List[int], dt: float) -> None:
        super().__init__("jerk_hist")
        self.bins = bins
        self.dt = dt

    def __call__(self, scene_df: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        accels: np.ndarray = np.linalg.norm(scene_df[["ax", "ay"]], axis=1)
        jerk: np.ndarray = (
            arr_utils.agent_aware_diff(accels, scene_df.index.get_level_values(0))
            / self.dt
        )

        return np.histogram(jerk, bins=self.bins)


def calc_stats(
    positions: Tensor, heading: Tensor, dt: float, bins: Dict[str, Tensor]
) -> Dict[str, Tensor]:
    """Calculate scene statistics for a simulated scene.

    Args:
        positions (Tensor): N x T x 2 tensor of agent positions (in world coordinates).
        heading (Tensor): N x T x 1 tensor of agent headings (in world coordinates).
        dt (float): The data's delta timestep.
        bins (Dict[str, Tensor]): A mapping from statistic name to a Tensor of bin edges.

    Returns:
        Dict[str, Tensor]: A mapping of value names to histograms.
    """

    velocity: Tensor = (
        torch.diff(positions, dim=1, prepend=positions[:, [1]] - positions[:, [0]]) / dt
    )
    velocity_norm: Tensor = torch.linalg.vector_norm(velocity, dim=-1)

    accel: Tensor = (
        torch.diff(positions, dim=1, prepend=velocity[:, [1]] - velocity[:, [0]]) / dt
    )
    accel_norm: Tensor = torch.linalg.vector_norm(accel, dim=-1)

    lon_acc: Tensor = accel_norm * torch.cos(heading.squeeze(-1))
    lat_acc: Tensor = accel_norm * torch.sin(heading.squeeze(-1))

    jerk: Tensor = (
        torch.diff(accel_norm, dim=1, prepend=accel_norm[:, [1]] - accel_norm[:, [0]])
        / dt
    )

    return {
        "velocity": torch.histogram(velocity_norm, bins["velocity"]),
        "lon_accel": torch.histogram(lon_acc, bins["lon_accel"]),
        "lat_accel": torch.histogram(lat_acc, bins["lat_accel"]),
        "jerk": torch.histogram(jerk, bins["jerk"]),
    }



from collections import defaultdict
from typing import Dict

import cv2
import numpy as np
from l5kit.data.map_api import InterpolationMethod
from l5kit.geometry import transform_points
from l5kit.rasterization.semantic_rasterizer import (
    CV2_SUB_VALUES,
    INTERPOLATION_POINTS,
    RasterEls,
    SemanticRasterizer,
    cv2_subpixel,
)


def indices_in_bounds(
    center: np.ndarray, bounds: np.ndarray, half_extent: float
) -> np.ndarray:
    """
    Get indices of elements for which the bounding box described by bounds intersects the one defined around
    center (square with side 2*half_side)

    Args:
        center (float): XY of the center
        bounds (np.ndarray): array of shape Nx2x2 [[x_min,y_min],[x_max, y_max]]
        half_extent (float): half the side of the bounding box centered around center

    Returns:
        np.ndarray: indices of elements inside radius from center
    """
    return np.arange(bounds.shape[0], dtype=np.long)


class MapSemanticRasterizer(SemanticRasterizer):
    def render_semantic_map(
        self, center_in_world: np.ndarray, raster_from_world: np.ndarray
    ) -> np.ndarray:
        """Renders the semantic map at given x,y coordinates.

        Args:
            center_in_world (np.ndarray): XY of the image center in world ref system
            raster_from_world (np.ndarray):
        Returns:
            np.ndarray: RGB raster

        """
        lane_area_img: np.ndarray = np.zeros(
            shape=(self.raster_size[1], self.raster_size[0], 3), dtype=np.uint8
        )
        lane_line_img: np.ndarray = np.zeros(
            shape=(self.raster_size[1], self.raster_size[0], 3), dtype=np.uint8
        )
        ped_area_img: np.ndarray = np.zeros(
            shape=(self.raster_size[1], self.raster_size[0], 3), dtype=np.uint8
        )

        # filter using half a radius from the center
        raster_radius = float(np.linalg.norm(self.raster_size * self.pixel_size)) / 2

        # get all lanes as interpolation so that we can transform them all together
        lane_indices = indices_in_bounds(
            center_in_world, self.mapAPI.bounds_info["lanes"]["bounds"], raster_radius
        )
        lanes_mask: Dict[str, np.ndarray] = defaultdict(
            lambda: np.zeros(len(lane_indices) * 2, dtype=np.bool)
        )
        lanes_area = np.zeros((len(lane_indices) * 2, INTERPOLATION_POINTS, 2))

        for idx, lane_idx in enumerate(lane_indices):
            lane_idx = self.mapAPI.bounds_info["lanes"]["ids"][lane_idx]

            # interpolate over polyline to always have the same number of points
            lane_coords = self.mapAPI.get_lane_as_interpolation(
                lane_idx, INTERPOLATION_POINTS, InterpolationMethod.INTER_ENSURE_LEN
            )
            lanes_area[idx * 2] = lane_coords["xyz_left"][:, :2]
            lanes_area[idx * 2 + 1] = lane_coords["xyz_right"][::-1, :2]

            lanes_mask[RasterEls.LANE_NOTL.name][idx * 2 : idx * 2 + 2] = True

        if len(lanes_area):
            lanes_area = cv2_subpixel(
                transform_points(lanes_area.reshape((-1, 2)), raster_from_world)
            )

            for lane_area in lanes_area.reshape((-1, INTERPOLATION_POINTS * 2, 2)):
                # need to for-loop otherwise some of them are empty
                cv2.fillPoly(lane_area_img, [lane_area], (255, 0, 0), **CV2_SUB_VALUES)

            lanes_area = lanes_area.reshape((-1, INTERPOLATION_POINTS, 2))
            for (
                name,
                mask,
            ) in lanes_mask.items():  # draw each type of lane with its own color
                cv2.polylines(
                    lane_line_img,
                    lanes_area[mask],
                    False,
                    (0, 255, 0),
                    **CV2_SUB_VALUES
                )

        # plot crosswalks
        crosswalks = []
        for idx in indices_in_bounds(
            center_in_world,
            self.mapAPI.bounds_info["crosswalks"]["bounds"],
            raster_radius,
        ):
            crosswalk = self.mapAPI.get_crosswalk_coords(
                self.mapAPI.bounds_info["crosswalks"]["ids"][idx]
            )
            xy_cross = cv2_subpixel(
                transform_points(crosswalk["xyz"][:, :2], raster_from_world)
            )
            crosswalks.append(xy_cross)

        cv2.fillPoly(ped_area_img, crosswalks, (0, 0, 255), **CV2_SUB_VALUES)

        map_img: np.ndarray = (lane_area_img + lane_line_img + ped_area_img).astype(
            np.float32
        ) / 255
        return map_img.transpose(2, 0, 1)



import warnings
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd
from nuscenes.eval.prediction.splits import NUM_IN_TRAIN_VAL
from nuscenes.map_expansion import arcline_path_utils
from nuscenes.map_expansion.map_api import NuScenesMap, locations
from nuscenes.nuscenes import NuScenes
from nuscenes.utils.splits import create_splits_scenes
from scipy.spatial.distance import cdist
from tqdm import tqdm

from trajdata.caching import EnvCache, SceneCache
from trajdata.data_structures.agent import (
    Agent,
    AgentMetadata,
    AgentType,
    FixedExtent,
    VariableExtent,
)
from trajdata.data_structures.environment import EnvMetadata
from trajdata.data_structures.scene_metadata import Scene, SceneMetadata
from trajdata.data_structures.scene_tag import SceneTag
from trajdata.dataset_specific.nusc import nusc_utils
from trajdata.dataset_specific.raw_dataset import RawDataset
from trajdata.dataset_specific.scene_records import NuscSceneRecord
from trajdata.maps import RasterizedMap, RasterizedMapMetadata, map_utils
from trajdata.proto.vectorized_map_pb2 import (
    MapElement,
    PedCrosswalk,
    PedWalkway,
    Polyline,
    RoadArea,
    RoadLane,
    VectorizedMap,
)


class NuscDataset(RawDataset):
    def compute_metadata(self, env_name: str, data_dir: str) -> EnvMetadata:
        # We're using the nuScenes prediction challenge split here.
        # See https://github.com/nutonomy/nuscenes-devkit/blob/master/python-sdk/nuscenes/eval/prediction/splits.py
        # for full details on how the splits are obtained below.
        all_scene_splits: Dict[str, List[str]] = create_splits_scenes()

        train_scenes: List[str] = deepcopy(all_scene_splits["train"])
        all_scene_splits["train"] = train_scenes[NUM_IN_TRAIN_VAL:]
        all_scene_splits["train_val"] = train_scenes[:NUM_IN_TRAIN_VAL]

        if env_name == "nusc_trainval":
            nusc_scene_splits: Dict[str, List[str]] = {
                k: all_scene_splits[k] for k in ["train", "train_val", "val"]
            }

            # nuScenes possibilities are the Cartesian product of these
            dataset_parts: List[Tuple[str, ...]] = [
                ("train", "train_val", "val"),
                ("boston", "singapore"),
            ]
        elif env_name == "nusc_test":
            nusc_scene_splits: Dict[str, List[str]] = {
                k: all_scene_splits[k] for k in ["test"]
            }

            # nuScenes possibilities are the Cartesian product of these
            dataset_parts: List[Tuple[str, ...]] = [
                ("test",),
                ("boston", "singapore"),
            ]

            warnings.warn("Beware, nusc_test has no annotations!")
        elif env_name == "nusc_mini":
            nusc_scene_splits: Dict[str, List[str]] = {
                k: all_scene_splits[k] for k in ["mini_train", "mini_val"]
            }

            # nuScenes possibilities are the Cartesian product of these
            dataset_parts: List[Tuple[str, ...]] = [
                ("mini_train", "mini_val"),
                ("boston", "singapore"),
            ]

        # Inverting the dict from above, associating every scene with its data split.
        nusc_scene_split_map: Dict[str, str] = {
            v_elem: k for k, v in nusc_scene_splits.items() for v_elem in v
        }

        return EnvMetadata(
            name=env_name,
            data_dir=data_dir,
            dt=nusc_utils.NUSC_DT,
            parts=dataset_parts,
            scene_split_map=nusc_scene_split_map,
        )

    def load_dataset_obj(self, verbose: bool = False) -> None:
        if verbose:
            print(f"Loading {self.name} dataset...", flush=True)

        if self.name == "nusc_mini":
            version_str = "v1.0-mini"
        elif self.name == "nusc_trainval":
            version_str = "v1.0-trainval"
        elif self.name == "nusc_test":
            version_str = "v1.0-test"

        self.dataset_obj = NuScenes(
            version=version_str, dataroot=self.metadata.data_dir
        )

    def _get_matching_scenes_from_obj(
        self,
        scene_tag: SceneTag,
        scene_desc_contains: Optional[List[str]],
        env_cache: EnvCache,
    ) -> List[SceneMetadata]:
        all_scenes_list: List[NuscSceneRecord] = list()

        scenes_list: List[SceneMetadata] = list()
        for idx, scene_record in enumerate(self.dataset_obj.scene):
            scene_name: str = scene_record["name"]
            scene_desc: str = scene_record["description"].lower()
            scene_location: str = self.dataset_obj.get(
                "log", scene_record["log_token"]
            )["location"]
            scene_split: str = self.metadata.scene_split_map[scene_name]
            scene_length: int = scene_record["nbr_samples"]

            # Saving all scene records for later caching.
            all_scenes_list.append(
                NuscSceneRecord(
                    scene_name, scene_location, scene_length, scene_desc, idx
                )
            )

            if scene_location.split("-")[0] in scene_tag and scene_split in scene_tag:
                if scene_desc_contains is not None and not any(
                    desc_query in scene_desc for desc_query in scene_desc_contains
                ):
                    continue

                scene_metadata = SceneMetadata(
                    env_name=self.metadata.name,
                    name=scene_name,
                    dt=self.metadata.dt,
                    raw_data_idx=idx,
                )
                scenes_list.append(scene_metadata)

        self.cache_all_scenes_list(env_cache, all_scenes_list)
        return scenes_list

    def _get_matching_scenes_from_cache(
        self,
        scene_tag: SceneTag,
        scene_desc_contains: Optional[List[str]],
        env_cache: EnvCache,
    ) -> List[Scene]:
        all_scenes_list: List[NuscSceneRecord] = env_cache.load_env_scenes_list(
            self.name
        )

        scenes_list: List[SceneMetadata] = list()
        for scene_record in all_scenes_list:
            (
                scene_name,
                scene_location,
                scene_length,
                scene_desc,
                data_idx,
            ) = scene_record
            scene_split: str = self.metadata.scene_split_map[scene_name]

            if scene_location.split("-")[0] in scene_tag and scene_split in scene_tag:
                if scene_desc_contains is not None and not any(
                    desc_query in scene_desc for desc_query in scene_desc_contains
                ):
                    continue

                scene_metadata = Scene(
                    self.metadata,
                    scene_name,
                    scene_location,
                    scene_split,
                    scene_length,
                    data_idx,
                    None,  # This isn't used if everything is already cached.
                    scene_desc,
                )
                scenes_list.append(scene_metadata)

        return scenes_list

    def get_scene(self, scene_info: SceneMetadata) -> Scene:
        _, _, _, data_idx = scene_info

        scene_record = self.dataset_obj.scene[data_idx]
        scene_name: str = scene_record["name"]
        scene_desc: str = scene_record["description"].lower()
        scene_location: str = self.dataset_obj.get("log", scene_record["log_token"])[
            "location"
        ]
        scene_split: str = self.metadata.scene_split_map[scene_name]
        scene_length: int = scene_record["nbr_samples"]

        return Scene(
            self.metadata,
            scene_name,
            scene_location,
            scene_split,
            scene_length,
            data_idx,
            scene_record,
            scene_desc,
        )

    def get_agent_info(
        self, scene: Scene, cache_path: Path, cache_class: Type[SceneCache]
    ) -> Tuple[List[AgentMetadata], List[List[AgentMetadata]]]:
        ego_agent_info: AgentMetadata = AgentMetadata(
            name="ego",
            agent_type=AgentType.VEHICLE,
            first_timestep=0,
            last_timestep=scene.length_timesteps - 1,
            extent=FixedExtent(length=4.084, width=1.730, height=1.562),
        )

        agent_presence: List[List[AgentMetadata]] = [
            [ego_agent_info] for _ in range(scene.length_timesteps)
        ]

        agent_data_list: List[pd.DataFrame] = list()
        existing_agents: Dict[str, AgentMetadata] = dict()

        all_frames: List[Dict[str, Union[str, int]]] = list(
            nusc_utils.frame_iterator(self.dataset_obj, scene)
        )
        frame_idx_dict: Dict[str, int] = {
            frame_dict["token"]: idx for idx, frame_dict in enumerate(all_frames)
        }
        for frame_idx, frame_info in enumerate(all_frames):
            for agent_info in nusc_utils.agent_iterator(self.dataset_obj, frame_info):
                if agent_info["instance_token"] in existing_agents:
                    continue

                if not agent_info["next"]:
                    # There are some agents with only a single detection to them, we don't care about these.
                    continue

                agent: Agent = nusc_utils.agg_agent_data(
                    self.dataset_obj, agent_info, frame_idx, frame_idx_dict
                )

                for scene_ts in range(
                    agent.metadata.first_timestep, agent.metadata.last_timestep + 1
                ):
                    agent_presence[scene_ts].append(agent.metadata)

                existing_agents[agent.name] = agent.metadata

                agent_data_list.append(agent.data)

        ego_agent: Agent = nusc_utils.agg_ego_data(self.dataset_obj, scene)
        agent_data_list.append(ego_agent.data)

        agent_list: List[AgentMetadata] = [ego_agent_info] + list(
            existing_agents.values()
        )

        cache_class.save_agent_data(pd.concat(agent_data_list), cache_path, scene)

        return agent_list, agent_presence

    def extract_lane_and_edges(
        self, nusc_map: NuScenesMap, lane_record
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        # Getting the bounding polygon vertices.
        lane_polygon_obj = nusc_map.get("polygon", lane_record["polygon_token"])
        polygon_nodes = [
            nusc_map.get("node", node_token)
            for node_token in lane_polygon_obj["exterior_node_tokens"]
        ]
        polygon_pts: np.ndarray = np.array(
            [(node["x"], node["y"]) for node in polygon_nodes]
        )

        # Getting the lane center's points.
        curr_lane = nusc_map.arcline_path_3.get(lane_record["token"], [])
        lane_midline: np.ndarray = np.array(
            arcline_path_utils.discretize_lane(curr_lane, resolution_meters=0.5)
        )[:, :2]

        # For some reason, nuScenes duplicates a few entries
        # (likely how they're building their arcline representation).
        # We delete those duplicate entries here.
        duplicate_check: np.ndarray = np.where(
            np.linalg.norm(np.diff(lane_midline, axis=0, prepend=0), axis=1) < 1e-10
        )[0]
        if duplicate_check.size > 0:
            lane_midline = np.delete(lane_midline, duplicate_check, axis=0)

        # Computing the closest lane center point to each bounding polygon vertex.
        closest_midlane_pt: np.ndarray = np.argmin(
            cdist(polygon_pts, lane_midline), axis=1
        )
        # Computing the local direction of the lane at each lane center point.
        direction_vectors: np.ndarray = np.diff(
            lane_midline,
            axis=0,
            prepend=lane_midline[[0]] - (lane_midline[[1]] - lane_midline[[0]]),
        )

        # Selecting the direction vectors at the closest lane center point per polygon vertex.
        local_dir_vecs: np.ndarray = direction_vectors[closest_midlane_pt]
        # Calculating the vectors from the the closest lane center point per polygon vertex to the polygon vertex.
        origin_to_polygon_vecs: np.ndarray = (
            polygon_pts - lane_midline[closest_midlane_pt]
        )

        # Computing the perpendicular dot product.
        # See https://www.xarg.org/book/linear-algebra/2d-perp-product/
        # If perp_dot_product < 0, then the associated polygon vertex is
        # on the right edge of the lane.
        perp_dot_product: np.ndarray = (
            local_dir_vecs[:, 0] * origin_to_polygon_vecs[:, 1]
            - local_dir_vecs[:, 1] * origin_to_polygon_vecs[:, 0]
        )

        # Determining which indices are on the right of the lane center.
        on_right: np.ndarray = perp_dot_product < 0
        # Determining the boundary between the left/right polygon vertices
        # (they will be together in blocks due to the ordering of the polygon vertices).
        idx_changes: int = np.where(np.roll(on_right, 1) < on_right)[0].item()

        if idx_changes > 0:
            # If the block of left/right points spreads across the bounds of the array,
            # roll it until the boundary between left/right points is at index 0.
            # This is important so that the following index selection orders points
            # without jumps.
            polygon_pts = np.roll(polygon_pts, shift=-idx_changes, axis=0)
            on_right = np.roll(on_right, shift=-idx_changes)

        left_pts: np.ndarray = polygon_pts[~on_right]
        right_pts: np.ndarray = polygon_pts[on_right]

        # Final ordering check, ensuring that the beginning of left_pts/right_pts
        # matches the beginning of the lane.
        left_order_correct: bool = np.linalg.norm(
            left_pts[0] - lane_midline[0]
        ) < np.linalg.norm(left_pts[0] - lane_midline[-1])
        right_order_correct: bool = np.linalg.norm(
            right_pts[0] - lane_midline[0]
        ) < np.linalg.norm(right_pts[0] - lane_midline[-1])

        # Reversing left_pts/right_pts in case their first index is
        # at the end of the lane.
        if not left_order_correct:
            left_pts = left_pts[::-1]
        if not right_order_correct:
            right_pts = right_pts[::-1]

        # Ensuring that left and right have the same number of points.
        # This is necessary, not for data storage but for later rasterization.
        if left_pts.shape[0] < right_pts.shape[0]:
            left_pts = map_utils.interpolate(left_pts, right_pts.shape[0])
        elif right_pts.shape[0] < left_pts.shape[0]:
            right_pts = map_utils.interpolate(right_pts, left_pts.shape[0])

        return (
            lane_midline,
            left_pts,
            right_pts,
        )

    def extract_area(self, nusc_map: NuScenesMap, area_record) -> np.ndarray:
        token_key: str
        if "exterior_node_tokens" in area_record:
            token_key = "exterior_node_tokens"
        elif "node_tokens" in area_record:
            token_key = "node_tokens"

        polygon_nodes = [
            nusc_map.get("node", node_token) for node_token in area_record[token_key]
        ]

        return np.array([(node["x"], node["y"]) for node in polygon_nodes])

    def extract_vectorized(self, nusc_map: NuScenesMap) -> VectorizedMap:
        vec_map = VectorizedMap()

        # Setting the map bounds.
        vec_map.max_pt.x, vec_map.max_pt.y, vec_map.max_pt.z = (
            nusc_map.explorer.canvas_max_x,
            nusc_map.explorer.canvas_max_y,
            0.0,
        )
        vec_map.min_pt.x, vec_map.min_pt.y, vec_map.min_pt.z = (
            nusc_map.explorer.canvas_min_x,
            nusc_map.explorer.canvas_min_y,
            0.0,
        )

        overall_pbar = tqdm(
            total=len(nusc_map.lane)
            + len(nusc_map.drivable_area[0]["polygon_tokens"])
            + len(nusc_map.ped_crossing)
            + len(nusc_map.walkway),
            desc=f"Getting {nusc_map.map_name} Elements",
            position=1,
            leave=False,
        )

        for lane_record in nusc_map.lane:
            center_pts, left_pts, right_pts = self.extract_lane_and_edges(
                nusc_map, lane_record
            )

            lane_record_token: str = lane_record["token"]

            # Adding the element to the map.
            new_element: MapElement = vec_map.elements.add()
            new_element.id = lane_record_token.encode()

            new_lane: RoadLane = new_element.road_lane
            map_utils.populate_lane_polylines(new_lane, center_pts, left_pts, right_pts)

            new_lane.entry_lanes.extend(
                lane_id.encode()
                for lane_id in nusc_map.get_incoming_lane_ids(lane_record_token)
            )
            new_lane.exit_lanes.extend(
                lane_id.encode()
                for lane_id in nusc_map.get_outgoing_lane_ids(lane_record_token)
            )

            # new_lane.adjacent_lanes_left.append(
            #     l5_lane.adjacent_lane_change_left.id
            # )
            # new_lane.adjacent_lanes_right.append(
            #     l5_lane.adjacent_lane_change_right.id
            # )

            overall_pbar.update()

        for polygon_token in nusc_map.drivable_area[0]["polygon_tokens"]:
            polygon_record = nusc_map.get("polygon", polygon_token)
            polygon_pts = self.extract_area(nusc_map, polygon_record)

            # Adding the element to the map.
            new_element: MapElement = vec_map.elements.add()
            new_element.id = lane_record["token"].encode()

            new_area: RoadArea = new_element.road_area
            map_utils.populate_polygon(new_area.exterior_polygon, polygon_pts)

            for hole in polygon_record["holes"]:
                polygon_pts = self.extract_area(nusc_map, hole)
                new_hole: Polyline = new_area.interior_holes.add()
                map_utils.populate_polygon(new_hole, polygon_pts)

            overall_pbar.update()

        for ped_area_record in nusc_map.ped_crossing:
            polygon_pts = self.extract_area(nusc_map, ped_area_record)

            # Adding the element to the map.
            new_element: MapElement = vec_map.elements.add()
            new_element.id = ped_area_record["token"].encode()

            new_crosswalk: PedCrosswalk = new_element.ped_crosswalk
            map_utils.populate_polygon(new_crosswalk.polygon, polygon_pts)

            overall_pbar.update()

        for ped_area_record in nusc_map.walkway:
            polygon_pts = self.extract_area(nusc_map, ped_area_record)

            # Adding the element to the map.
            new_element: MapElement = vec_map.elements.add()
            new_element.id = ped_area_record["token"].encode()

            new_walkway: PedWalkway = new_element.ped_walkway
            map_utils.populate_polygon(new_walkway.polygon, polygon_pts)

            overall_pbar.update()

        overall_pbar.close()

        return vec_map

    def cache_map(
        self,
        map_name: str,
        cache_path: Path,
        map_cache_class: Type[SceneCache],
        map_params: Dict[str, Any],
    ) -> None:
        resolution: float = map_params["px_per_m"]

        nusc_map: NuScenesMap = NuScenesMap(
            dataroot=self.metadata.data_dir, map_name=map_name
        )

        if map_params.get("original_format", False):
            warnings.warn(
                "Using a dataset's original map format is deprecated, and will be removed in the next version of trajdata!",
                FutureWarning,
            )

            width_m, height_m = nusc_map.canvas_edge
            height_px, width_px = round(height_m * resolution), round(
                width_m * resolution
            )

            def layer_fn(layer_name: str) -> np.ndarray:
                # Getting rid of the channels dim by accessing index [0]
                return nusc_map.get_map_mask(
                    patch_box=None,
                    patch_angle=0,
                    layer_names=[layer_name],
                    canvas_size=(height_px, width_px),
                )[0].astype(np.bool)

            map_from_world: np.ndarray = np.array(
                [[resolution, 0.0, 0.0], [0.0, resolution, 0.0], [0.0, 0.0, 1.0]]
            )

            layer_names: List[str] = [
                "lane",
                "road_segment",
                "drivable_area",
                "road_divider",
                "lane_divider",
                "ped_crossing",
                "walkway",
            ]
            map_info: RasterizedMapMetadata = RasterizedMapMetadata(
                name=map_name,
                shape=(len(layer_names), height_px, width_px),
                layers=layer_names,
                layer_rgb_groups=([0, 1, 2], [3, 4], [5, 6]),
                resolution=resolution,
                map_from_world=map_from_world,
            )

            map_cache_class.cache_map_layers(
                cache_path, VectorizedMap(), map_info, layer_fn, self.name
            )
        else:
            vectorized_map: VectorizedMap = self.extract_vectorized(nusc_map)

            pbar_kwargs = {"position": 2, "leave": False}
            map_data, map_from_world = map_utils.rasterize_map(
                vectorized_map, resolution, **pbar_kwargs
            )

            rasterized_map_info: RasterizedMapMetadata = RasterizedMapMetadata(
                name=map_name,
                shape=map_data.shape,
                layers=["drivable_area", "lane_divider", "ped_area"],
                layer_rgb_groups=([0], [1], [2]),
                resolution=resolution,
                map_from_world=map_from_world,
            )

            rasterized_map_obj: RasterizedMap = RasterizedMap(
                rasterized_map_info, map_data
            )

            map_cache_class.cache_map(
                cache_path, vectorized_map, rasterized_map_obj, self.name
            )

    def cache_maps(
        self,
        cache_path: Path,
        map_cache_class: Type[SceneCache],
        map_params: Dict[str, Any],
    ) -> None:
        """
        Stores rasterized maps to disk for later retrieval.

        Below are the map origins (south western corner, in [lat, lon]) for each of
        the 4 maps in nuScenes:
            boston-seaport: [42.336849169438615, -71.05785369873047]
            singapore-onenorth: [1.2882100868743724, 103.78475189208984]
            singapore-hollandvillage: [1.2993652317780957, 103.78217697143555]
            singapore-queenstown: [1.2782562240223188, 103.76741409301758]

        The dimensions of the maps are as follows ([width, height] in meters). They
        can also be found in nusc_utils.py
            singapore-onenorth:       [1585.6, 2025.0]
            singapore-hollandvillage: [2808.3, 2922.9]
            singapore-queenstown:     [3228.6, 3687.1]
            boston-seaport:           [2979.5, 2118.1]
        The rasterized semantic maps published with nuScenes v1.0 have a scale of 10px/m,
        hence the above numbers are the image dimensions divided by 10.

        nuScenes uses the same WGS 84 Web Mercator (EPSG:3857) projection as Google Maps/Earth.
        """
        for map_name in tqdm(
            locations,
            desc=f"Caching {self.name} Maps at {map_params['px_per_m']:.2f} px/m",
            position=0,
        ):
            self.cache_map(map_name, cache_path, map_cache_class, map_params)



from typing import List, Optional, Set

import numpy as np
import pandas as pd

from trajdata import filtering
from trajdata.caching import SceneCache
from trajdata.data_structures.agent import Agent, AgentMetadata, AgentType
from trajdata.data_structures.scene_metadata import Scene


class SceneTime:
    """Holds the data for a particular scene at a particular timestep."""

    def __init__(
        self,
        scene: Scene,
        scene_ts: int,
        agents: List[AgentMetadata],
        cache: SceneCache,
    ) -> None:
        self.scene = scene
        self.ts = scene_ts
        self.agents = agents
        self.cache = cache

    @classmethod
    def from_cache(
        cls,
        scene: Scene,
        scene_ts: int,
        cache: SceneCache,
        only_types: Optional[Set[AgentType]] = None,
        no_types: Optional[Set[AgentType]] = None,
    ):
        agents_present: List[AgentMetadata] = scene.agent_presence[scene_ts]
        filtered_agents: List[AgentMetadata] = filtering.agent_types(
            agents_present, no_types, only_types
        )

        return cls(scene, scene_ts, filtered_agents, cache)

    def get_agent_distances_to(self, agent: Agent) -> np.ndarray:
        agent_pos: np.ndarray = self.cache.get_state(agent.name, self.ts)[:2]
        nb_pos: np.ndarray = np.stack(
            [self.cache.get_state(nb.name, self.ts)[:2] for nb in self.agents]
        )

        return np.linalg.norm(nb_pos - agent_pos, axis=1)


class SceneTimeAgent:
    """Holds the data for a particular agent in a scene at a particular timestep."""

    def __init__(
        self,
        scene: Scene,
        scene_ts: int,
        agents: List[AgentMetadata],
        agent: AgentMetadata,
        cache: SceneCache,
        robot: Optional[AgentMetadata] = None,
    ) -> None:
        self.scene = scene
        self.ts = scene_ts
        self.agents = agents
        self.agent = agent
        self.cache = cache
        self.robot = robot

    @classmethod
    def from_cache(
        cls,
        scene: Scene,
        scene_ts: int,
        agent_id: str,
        cache: SceneCache,
        only_types: Optional[Set[AgentType]] = None,
        no_types: Optional[Set[AgentType]] = None,
        incl_robot_future: bool = False,
    ):
        agents_present: List[AgentMetadata] = scene.agent_presence[scene_ts]
        filtered_agents: List[AgentMetadata] = filtering.agent_types(
            agents_present, no_types, only_types
        )

        agent_metadata = next((a for a in filtered_agents if a.name == agent_id), None)

        if incl_robot_future:
            ego_metadata = next((a for a in filtered_agents if a.name == "ego"), None)

            return cls(
                scene,
                scene_ts,
                agents=filtered_agents,
                agent=agent_metadata,
                cache=cache,
                robot=ego_metadata,
            )
        else:
            return cls(
                scene,
                scene_ts,
                agents=filtered_agents,
                agent=agent_metadata,
                cache=cache,
            )

    # @profile
    def get_agent_distances_to(self, agent_info: AgentMetadata) -> np.ndarray:
        agent_pos: np.ndarray = self.cache.get_state(agent_info.name, self.ts)[:2]

        curr_poses: np.ndarray = self.cache.get_states(
            [a.name for a in self.agents], self.ts
        )[:, :2]
        return np.linalg.norm(curr_poses - agent_pos, axis=1)



from pathlib import Path
from typing import Dict, List, Optional, Type

import numpy as np
from torch.utils.data import Dataset

from trajdata.caching import EnvCache, SceneCache
from trajdata.data_structures import Scene, SceneMetadata
from trajdata.utils import agent_utils
from trajdata.utils.env_utils import get_raw_dataset


def scene_paths_collate_fn(filled_scenes: List) -> List:
    return filled_scenes


class ParallelDatasetPreprocessor(Dataset):
    def __init__(
        self,
        scene_info_list: List[SceneMetadata],
        envs_dir_dict: Dict[str, str],
        env_cache_path: str,
        desired_dt: Optional[float],
        cache_class: Type[SceneCache],
        rebuild_cache: bool,
    ) -> None:
        self.env_cache_path = np.array(env_cache_path).astype(np.string_)
        self.desired_dt = desired_dt
        self.cache_class = cache_class
        self.rebuild_cache = rebuild_cache

        env_names: List[str] = list(envs_dir_dict.keys())
        scene_names: List[str] = [scene_info.name for scene_info in scene_info_list]

        self.scene_idxs = np.array(
            [scene_info.raw_data_idx for scene_info in scene_info_list], dtype=int
        )
        self.env_name_idxs = np.array(
            [env_names.index(scene_info.env_name) for scene_info in scene_info_list],
            dtype=int,
        )
        self.scene_name_idxs = np.array(
            [scene_names.index(scene_info.name) for scene_info in scene_info_list],
            dtype=int,
        )

        self.env_names_arr = np.array(env_names).astype(np.string_)
        self.scene_names_arr = np.array(scene_names).astype(np.string_)
        self.data_dir_arr = np.array(list(envs_dir_dict.values())).astype(np.string_)

        self.data_len: int = len(scene_info_list)

    def __len__(self) -> int:
        return self.data_len

    def __getitem__(self, idx: int) -> str:
        env_cache_path: Path = Path(str(self.env_cache_path, encoding="utf-8"))
        env_cache: EnvCache = EnvCache(env_cache_path)

        env_idx: int = self.env_name_idxs[idx]
        scene_idx: int = self.scene_name_idxs[idx]

        env_name: str = str(self.env_names_arr[env_idx], encoding="utf-8")
        raw_dataset = get_raw_dataset(
            env_name, str(self.data_dir_arr[env_idx], encoding="utf-8")
        )

        scene_name: str = str(self.scene_names_arr[scene_idx], encoding="utf-8")

        scene_info = SceneMetadata(
            env_name, scene_name, raw_dataset.metadata.dt, self.scene_idxs[idx]
        )

        # Leaving verbose False here so that we don't spam
        # stdout with loading messages.
        raw_dataset.load_dataset_obj(verbose=False)
        scene: Scene = agent_utils.get_agent_data(
            scene_info,
            raw_dataset,
            env_cache,
            self.rebuild_cache,
            self.cache_class,
            self.desired_dt,
        )
        raw_dataset.del_dataset_obj()

        scene_path: Path = EnvCache.scene_metadata_path(
            env_cache.path, scene.env_name, scene.name, scene.dt
        )
        return str(scene_path)



from typing import Dict

import numpy as np
import pandas as pd


class SimMetric:
    def __init__(self, name: str) -> None:
        self.name = name

    def __call__(self, gt_df: pd.DataFrame, sim_df: pd.DataFrame) -> Dict[str, float]:
        raise NotImplementedError()


class ADE(SimMetric):
    def __init__(self) -> None:
        super().__init__("ade")

    def __call__(self, gt_df: pd.DataFrame, sim_df: pd.DataFrame) -> Dict[str, float]:
        err_df = pd.DataFrame(index=gt_df.index, columns=["error"])
        err_df["error"] = np.linalg.norm(gt_df[["x", "y"]] - sim_df[["x", "y"]], axis=1)
        return err_df.groupby("agent_id")["error"].mean().to_dict()


class FDE(SimMetric):
    def __init__(self) -> None:
        super().__init__("fde")

    def __call__(self, gt_df: pd.DataFrame, sim_df: pd.DataFrame) -> Dict[str, float]:
        err_df = pd.DataFrame(index=gt_df.index, columns=["error"])
        err_df["error"] = np.linalg.norm(gt_df[["x", "y"]] - sim_df[["x", "y"]], axis=1)
        return err_df.groupby("agent_id")["error"].last().to_dict()



import pickle
from multiprocessing import Manager, Pool
from typing import Callable, Iterable, List, Optional

from tqdm import tqdm


def parallel_apply(
    element_fn: Callable,
    element_list: Iterable,
    num_workers: int,
    desc: Optional[str] = None,
    disable: bool = False,
) -> List:
    return list(parallel_iapply(element_fn, element_list, num_workers, desc, disable))


def parallel_iapply(
    element_fn: Callable,
    element_list: Iterable,
    num_workers: int,
    desc: Optional[str] = None,
    disable: bool = False,
) -> Iterable:
    with Pool(processes=num_workers) as pool:
        for fn_output in tqdm(
            pool.imap(element_fn, element_list),
            desc=desc,
            total=len(element_list),
            disable=disable,
        ):
            yield fn_output


def pickle_objects(objs: List) -> List[bytes]:
    pickled_objs: List[bytes] = list()
    for obj in objs:
        pickled_objs.append(pickle.dumps(obj))

    return pickled_objs


class AsyncExecutor:
    def __init__(
        self, num_workers: int, total_jobs: int, desc: str, position: int, disable: bool
    ):
        self.pool = Pool(processes=num_workers)
        self.manager = Manager()
        self.results_queue = self.manager.Queue()

        self.pbar = tqdm(
            desc=desc, position=position, total=total_jobs, disable=disable
        )

    def error(self, err):
        raise err

    def prompt(self, results: List[bytes]):
        self.pbar.update(len(results))
        for result in results:
            self.results_queue.put(result)

    def schedule(self, function, args):
        return self.pool.apply_async(
            function, args, callback=self.prompt, error_callback=self.error
        )

    def wait(self):
        self.pool.close()
        self.pool.join()
        self.pbar.close()



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional

import torch
from torch import Tensor

from trajdata.data_structures.agent import AgentType
from trajdata.utils.arr_utils import PadDirection


@dataclass
class AgentBatch:
    data_idx: Tensor
    dt: Tensor
    agent_name: List[str]
    agent_type: Tensor
    curr_agent_state: Tensor
    agent_hist: Tensor
    agent_hist_extent: Tensor
    agent_hist_len: Tensor
    agent_fut: Tensor
    agent_fut_extent: Tensor
    agent_fut_len: Tensor
    num_neigh: Tensor
    neigh_types: Tensor
    neigh_hist: Tensor
    neigh_hist_extents: Tensor
    neigh_hist_len: Tensor
    neigh_fut: Tensor
    neigh_fut_extents: Tensor
    neigh_fut_len: Tensor
    robot_fut: Optional[Tensor]
    robot_fut_len: Optional[Tensor]
    maps: Optional[Tensor]
    maps_resolution: Optional[Tensor]
    rasters_from_world_tf: Optional[Tensor]
    agents_from_world_tf: Tensor
    scene_ids: Optional[List]
    history_pad_dir: PadDirection
    extras: Dict[str, Tensor]

    def to(self, device) -> None:
        excl_vals = {
            "data_idx",
            "agent_name",
            "agent_type",
            "agent_hist_len",
            "agent_fut_len",
            "neigh_hist_len",
            "neigh_fut_len",
            "neigh_types",
            "num_neigh",
            "robot_fut_len",
            "scene_ids",
            "history_pad_dir",
            "extras",
        }
        for val in vars(self).keys():
            tensor_val = getattr(self, val)
            if val not in excl_vals and tensor_val is not None:
                tensor_val: Tensor
                setattr(self, val, tensor_val.to(device, non_blocking=True))

        for key, val in self.extras.items():
            self.extras[key] = val.to(device, non_blocking=True)

    def agent_types(self) -> List[AgentType]:
        unique_types: Tensor = torch.unique(self.agent_type)
        return [AgentType(unique_type.item()) for unique_type in unique_types]

    def for_agent_type(self, agent_type: AgentType) -> AgentBatch:
        match_type = self.agent_type == agent_type
        return AgentBatch(
            data_idx=self.data_idx[match_type],
            dt=self.dt[match_type],
            agent_name=[
                name for idx, name in enumerate(self.agent_name) if match_type[idx]
            ],
            agent_type=agent_type.value,
            curr_agent_state=self.curr_agent_state[match_type],
            agent_hist=self.agent_hist[match_type],
            agent_hist_extent=self.agent_hist_extent[match_type],
            agent_hist_len=self.agent_hist_len[match_type],
            agent_fut=self.agent_fut[match_type],
            agent_fut_extent=self.agent_fut_extent[match_type],
            agent_fut_len=self.agent_fut_len[match_type],
            num_neigh=self.num_neigh[match_type],
            neigh_types=self.neigh_types[match_type],
            neigh_hist=self.neigh_hist[match_type],
            neigh_hist_extents=self.neigh_hist_extents[match_type],
            neigh_hist_len=self.neigh_hist_len[match_type],
            neigh_fut=self.neigh_fut[match_type],
            neigh_fut_extents=self.neigh_fut_extents[match_type],
            neigh_fut_len=self.neigh_fut_len[match_type],
            robot_fut=self.robot_fut[match_type]
            if self.robot_fut is not None
            else None,
            robot_fut_len=self.robot_fut_len[match_type]
            if self.robot_fut_len is not None
            else None,
            maps=self.maps[match_type] if self.maps is not None else None,
            maps_resolution=self.maps_resolution[match_type]
            if self.maps_resolution is not None
            else None,
            rasters_from_world_tf=self.rasters_from_world_tf[match_type]
            if self.rasters_from_world_tf is not None
            else None,
            agents_from_world_tf=self.agents_from_world_tf[match_type],
            scene_ids=[
                scene_id
                for idx, scene_id in enumerate(self.scene_ids)
                if match_type[idx]
            ],
            history_pad_dir=self.history_pad_dir,
            extras={key: val[match_type] for key, val in self.extras},
        )


@dataclass
class SceneBatch:
    data_idx: Tensor
    dt: Tensor
    num_agents: Tensor
    agent_type: Tensor
    centered_agent_state: Tensor
    agent_hist: Tensor
    agent_hist_extent: Tensor
    agent_hist_len: Tensor
    agent_fut: Tensor
    agent_fut_extent: Tensor
    agent_fut_len: Tensor
    robot_fut: Optional[Tensor]
    robot_fut_len: Optional[Tensor]
    maps: Optional[Tensor]
    maps_resolution: Optional[Tensor]
    rasters_from_world_tf: Optional[Tensor]
    centered_agent_from_world_tf: Tensor
    centered_world_from_agent_tf: Tensor
    scene_ids: Optional[List]
    history_pad_dir: PadDirection
    extras: Dict[str, Tensor]

    def to(self, device) -> None:
        excl_vals = {
            "history_pad_dir",
            "extras",
        }

        for val in vars(self).keys():
            tensor_val = getattr(self, val)
            if val not in excl_vals and tensor_val is not None:
                setattr(self, val, tensor_val.to(device))

        for key, val in self.extras.items():
            self.extras[key] = val.to(device)

    def agent_types(self) -> List[AgentType]:
        unique_types: Tensor = torch.unique(self.agent_type)
        return [AgentType(unique_type.item()) for unique_type in unique_types]

    def for_agent_type(self, agent_type: AgentType) -> SceneBatch:
        match_type = self.agent_type == agent_type
        return SceneBatch(
            data_idx=self.data_idx[match_type],
            dt=self.dt[match_type],
            num_agents=self.num_agents[match_type],
            agent_type=self.agent_type[match_type],
            centered_agent_state=self.centered_agent_state[match_type],
            agent_hist=self.agent_hist[match_type],
            agent_hist_extent=self.agent_hist_extent[match_type],
            agent_hist_len=self.agent_hist_len[match_type],
            agent_fut=self.agent_fut[match_type],
            agent_fut_extent=self.agent_fut_extent[match_type],
            agent_fut_len=self.agent_fut_len[match_type],
            robot_fut=self.robot_fut[match_type]
            if self.robot_fut is not None
            else None,
            robot_fut_len=self.robot_fut_len[match_type]
            if self.robot_fut_len is not None
            else None,
            maps=self.maps[match_type] if self.maps is not None else None,
            maps_resolution=self.maps_resolution[match_type]
            if self.maps_resolution is not None
            else None,
            rasters_from_world_tf=self.rasters_from_world_tf[match_type]
            if self.rasters_from_world_tf is not None
            else None,
            centered_agent_from_world_tf=self.centered_agent_from_world_tf[match_type],
            centered_world_from_agent_tf=self.centered_world_from_agent_tf[match_type],
            scene_ids=[
                scene_id
                for idx, scene_id in enumerate(self.scene_ids)
                if match_type[idx]
            ],
            history_pad_dir=self.history_pad_dir,
            extras={key: val[match_type] for key, val in self.extras},
        )



from dataclasses import dataclass
from enum import IntEnum

import numpy as np
import pandas as pd


class AgentType(IntEnum):
    UNKNOWN = 0
    VEHICLE = 1
    PEDESTRIAN = 2
    BICYCLE = 3
    MOTORCYCLE = 4


class Extent:
    def get_extents(self, start_ts: int, end_ts: int) -> np.ndarray:
        """Get the agent's extents within the specified scene timesteps.

        Args:
            start_ts (int): The first scene timestep to get extents for (inclusive)
            end_ts (int): The last scene timestep to get extents for (inclusive)

        Returns:
            np.ndarray: The extents as a (T, 3)-shaped ndarray (length, width, height)
        """
        raise NotImplementedError()


@dataclass
class FixedExtent(Extent):
    length: float
    width: float
    height: float

    def get_extents(self, start_ts: int, end_ts: int) -> np.ndarray:
        return np.repeat(
            np.array([[self.length, self.width, self.height]]),
            end_ts - start_ts + 1,
            axis=0,
        )


class VariableExtent(Extent):
    pass


class AgentMetadata:
    """Holds node metadata, e.g., name, type, but without the memory footprint of all the actual underlying scene data."""

    def __init__(
        self,
        name: str,
        agent_type: AgentType,
        first_timestep: int,
        last_timestep: int,
        extent: Extent,
    ) -> None:
        self.name = name
        self.type = agent_type
        self.first_timestep = first_timestep
        self.last_timestep = last_timestep
        self.extent = extent

    def __repr__(self) -> str:
        return "/".join([self.type.name, self.name])


class Agent:
    """Holds the data for a particular node."""

    def __init__(
        self,
        metadata: AgentMetadata,
        data: pd.DataFrame,
    ) -> None:
        self.name = metadata.name
        self.type = metadata.type
        self.metadata = metadata
        self.data = data



from collections import namedtuple
from typing import Any, List, Optional

from trajdata.data_structures.agent import AgentMetadata
from trajdata.data_structures.environment import EnvMetadata

# Holds scene metadata (e.g., name, dt), but without the memory
# footprint of all the actual underlying scene data.
SceneMetadata = namedtuple("SceneMetadata", ["env_name", "name", "dt", "raw_data_idx"])


class Scene:
    """Holds the data for a particular scene."""

    def __init__(
        self,
        env_metadata: EnvMetadata,
        name: str,
        location: str,
        data_split: str,
        length_timesteps: int,
        raw_data_idx: int,
        data_access_info: Any,
        description: Optional[str] = None,
        agents: Optional[List[AgentMetadata]] = None,
        agent_presence: Optional[List[List[AgentMetadata]]] = None,
    ) -> None:
        self.env_metadata = env_metadata
        self.env_name = env_metadata.name
        self.name = name
        self.location = location
        self.data_split = data_split
        self.dt = env_metadata.dt
        self.length_timesteps = length_timesteps
        self.raw_data_idx = raw_data_idx
        self.data_access_info = data_access_info
        self.description = description
        self.agents = agents
        self.agent_presence = agent_presence

    def length_seconds(self) -> float:
        return self.length_timesteps * self.dt

    def __repr__(self) -> str:
        return "/".join([self.env_name, self.name])

    def update_agent_info(
        self,
        new_agents: List[AgentMetadata],
        new_agent_presence: List[List[AgentMetadata]],
    ) -> None:
        self.agents = new_agents
        self.agent_presence = new_agent_presence

    def to_metadata(self) -> SceneMetadata:
        return SceneMetadata(self.env_name, self.name, self.dt, self.raw_data_idx)



from typing import Optional

import matplotlib.pyplot as plt
import torch
from matplotlib.axes import Axes
from torch import Tensor

from trajdata.data_structures.agent import AgentType
from trajdata.data_structures.batch import AgentBatch, SceneBatch
from trajdata.maps import RasterizedMap


def plot_agent_batch(
    batch: AgentBatch,
    batch_idx: int,
    ax: Optional[Axes] = None,
    show: bool = True,
    close: bool = True,
) -> None:
    if ax is None:
        _, ax = plt.subplots()

    agent_name: str = batch.agent_name[batch_idx]
    agent_type: AgentType = AgentType(batch.agent_type[batch_idx].item())
    ax.set_title(f"{str(agent_type)}/{agent_name}")

    history_xy: Tensor = batch.agent_hist[batch_idx].cpu()
    center_xy: Tensor = batch.agent_hist[batch_idx, -1, :2].cpu()
    future_xy: Tensor = batch.agent_fut[batch_idx, :, :2].cpu()

    if batch.maps is not None:
        agent_from_world_tf: Tensor = batch.agents_from_world_tf[batch_idx].cpu()
        world_from_raster_tf: Tensor = torch.linalg.inv(
            batch.rasters_from_world_tf[batch_idx].cpu()
        )

        agent_from_raster_tf: Tensor = agent_from_world_tf @ world_from_raster_tf

        patch_size: int = batch.maps[batch_idx].shape[-1]

        left_extent: float = (agent_from_raster_tf @ torch.tensor([0.0, 0.0, 1.0]))[
            0
        ].item()
        right_extent: float = (
            agent_from_raster_tf @ torch.tensor([patch_size, 0.0, 1.0])
        )[0].item()
        bottom_extent: float = (
            agent_from_raster_tf @ torch.tensor([0.0, patch_size, 1.0])
        )[1].item()
        top_extent: float = (agent_from_raster_tf @ torch.tensor([0.0, 0.0, 1.0]))[
            1
        ].item()

        ax.imshow(
            RasterizedMap.to_img(
                batch.maps[batch_idx].cpu(),
                # [[0], [1], [2]]
                # [[0, 1, 2], [3, 4], [5, 6]],
            ),
            extent=(
                left_extent,
                right_extent,
                bottom_extent,
                top_extent,
            ),
            alpha=0.3,
        )

    ax.plot(
        history_xy[..., 0],
        history_xy[..., 1],
        c="orange",
        ls="--",
        label="Agent History",
    )
    ax.quiver(
        history_xy[..., 0],
        history_xy[..., 1],
        history_xy[..., -1],
        history_xy[..., -2],
        color="k",
    )
    ax.plot(future_xy[..., 0], future_xy[..., 1], c="violet", label="Agent Future")
    ax.scatter(center_xy[0], center_xy[1], s=20, c="orangered", label="Agent Current")

    num_neigh = batch.num_neigh[batch_idx]
    if num_neigh > 0:
        neighbor_hist = batch.neigh_hist[batch_idx]
        neighbor_fut = batch.neigh_fut[batch_idx]

        ax.plot([], [], c="olive", ls="--", label="Neighbor History")
        for n in range(num_neigh):
            ax.plot(neighbor_hist[n, :, 0], neighbor_hist[n, :, 1], c="olive", ls="--")

        ax.plot([], [], c="darkgreen", label="Neighbor Future")
        for n in range(num_neigh):
            ax.plot(neighbor_fut[n, :, 0], neighbor_fut[n, :, 1], c="darkgreen")

        ax.scatter(
            neighbor_hist[:num_neigh, -1, 0],
            neighbor_hist[:num_neigh, -1, 1],
            s=20,
            c="gold",
            label="Neighbor Current",
        )

    if batch.robot_fut is not None and batch.robot_fut.shape[1] > 0:
        ax.plot(
            batch.robot_fut[batch_idx, 1:, 0],
            batch.robot_fut[batch_idx, 1:, 1],
            label="Ego Future",
            c="blue",
        )
        ax.scatter(
            batch.robot_fut[batch_idx, 0, 0],
            batch.robot_fut[batch_idx, 0, 1],
            s=20,
            c="blue",
            label="Ego Current",
        )

    ax.set_xlabel("x (m)")
    ax.set_ylabel("y (m)")

    ax.grid(False)
    ax.legend(loc="best", frameon=True)
    ax.axis("equal")

    if show:
        plt.show()

    if close:
        plt.close()


def plot_scene_batch(
    batch: SceneBatch,
    batch_idx: int,
    ax: Optional[Axes] = None,
    show: bool = True,
    close: bool = True,
) -> None:
    if ax is None:
        _, ax = plt.subplots()

    num_agents: int = batch.num_agents[batch_idx].item()

    history_xy: Tensor = batch.agent_hist[batch_idx].cpu()
    center_xy: Tensor = batch.agent_hist[batch_idx, ..., -1, :2].cpu()
    future_xy: Tensor = batch.agent_fut[batch_idx, ..., :2].cpu()

    if batch.maps is not None:
        centered_agent_id: int = 0
        agent_from_world_tf: Tensor = batch.centered_agent_from_world_tf[
            batch_idx
        ].cpu()
        world_from_raster_tf: Tensor = torch.linalg.inv(
            batch.rasters_from_world_tf[batch_idx, centered_agent_id].cpu()
        )

        agent_from_raster_tf: Tensor = agent_from_world_tf @ world_from_raster_tf

        patch_size: int = batch.maps[batch_idx, centered_agent_id].shape[-1]

        left_extent: float = (agent_from_raster_tf @ torch.tensor([0.0, 0.0, 1.0]))[
            0
        ].item()
        right_extent: float = (
            agent_from_raster_tf @ torch.tensor([patch_size, 0.0, 1.0])
        )[0].item()
        bottom_extent: float = (
            agent_from_raster_tf @ torch.tensor([0.0, patch_size, 1.0])
        )[1].item()
        top_extent: float = (agent_from_raster_tf @ torch.tensor([0.0, 0.0, 1.0]))[
            1
        ].item()

        ax.imshow(
            RasterizedMap.to_img(
                batch.maps[batch_idx, centered_agent_id].cpu(),
                # [[0], [1], [2]]
                # [[0, 1, 2], [3, 4], [5, 6]],
            ),
            extent=(
                left_extent,
                right_extent,
                bottom_extent,
                top_extent,
            ),
            alpha=0.3,
        )

    for agent_id in range(num_agents):
        ax.plot(
            history_xy[agent_id, ..., 0],
            history_xy[agent_id, ..., 1],
            c="orange",
            ls="--",
            label="Agent History" if agent_id == 0 else None,
        )
        ax.quiver(
            history_xy[agent_id, ..., 0],
            history_xy[agent_id, ..., 1],
            history_xy[agent_id, ..., -1],
            history_xy[agent_id, ..., -2],
            color="k",
        )
        ax.plot(
            future_xy[agent_id, ..., 0],
            future_xy[agent_id, ..., 1],
            c="violet",
            label="Agent Future" if agent_id == 0 else None,
        )
        ax.scatter(
            center_xy[agent_id, 0],
            center_xy[agent_id, 1],
            s=20,
            c="orangered",
            label="Agent Current" if agent_id == 0 else None,
        )

    if batch.robot_fut is not None and batch.robot_fut.shape[1] > 0:
        ax.plot(
            batch.robot_fut[batch_idx, 1:, 0],
            batch.robot_fut[batch_idx, 1:, 1],
            label="Ego Future",
            c="blue",
        )
        ax.scatter(
            batch.robot_fut[batch_idx, 0, 0],
            batch.robot_fut[batch_idx, 0, 1],
            s=20,
            c="blue",
            label="Ego Current",
        )

    ax.set_xlabel("x (m)")
    ax.set_ylabel("y (m)")

    ax.grid(False)
    ax.legend(loc="best", frameon=True)
    ax.axis("equal")

    if show:
        plt.show()

    if close:
        plt.close()



from math import ceil
from typing import List, Optional, Union

from trajdata.data_structures import AgentMetadata, Scene, SceneMetadata


def enforce_desired_dt(
    scene_info: Union[Scene, SceneMetadata],
    desired_dt: Optional[float],
    dry_run: bool = False,
) -> bool:
    """Enforces that a scene's data is at the desired frequency (specified by desired_dt
    if it's not None) through interpolation.

    Args:
        scene_info (Scene | SceneMetadata): The scene to interpolate to the desired data frequency.
        desired_dt (Optional[float]): The desired data timestep difference (in seconds).
        dry_run (bool): If True, only check if the scene meets the desired data frequency (without modifying scene_info). Defaults to False.

    Returns:
        bool: True if the scene was modified (or would be modified if dry_run=True), False otherwise.
    """
    if desired_dt is not None and scene_info.dt != desired_dt:
        if not dry_run and scene_info.dt > desired_dt:
            interpolate_scene_dt(scene_info, desired_dt)
        elif not dry_run and scene_info.dt < desired_dt:
            subsample_scene_dt(scene_info, desired_dt)
        return True

    return False


def interpolate_scene_dt(scene: Scene, desired_dt: float) -> None:
    dt_ratio: float = scene.dt / desired_dt
    if not dt_ratio.is_integer():
        raise ValueError(
            f"Cannot interpolate scene: {scene.dt} is not integer divisible by {desired_dt} for {str(scene)}"
        )

    dt_factor: int = int(dt_ratio)

    # E.g., the scene is currently at dt = 0.5s (2 Hz),
    # but we want desired_dt = 0.1s (10 Hz).
    scene.length_timesteps = (scene.length_timesteps - 1) * dt_factor + 1
    agent_presence: List[List[AgentMetadata]] = [
        [] for _ in range(scene.length_timesteps)
    ]
    for agent in scene.agents:
        agent.first_timestep *= dt_factor
        agent.last_timestep *= dt_factor

        for scene_ts in range(agent.first_timestep, agent.last_timestep + 1):
            agent_presence[scene_ts].append(agent)

    scene.update_agent_info(scene.agents, agent_presence)
    scene.dt = desired_dt
    # Note we do not touch scene_info.env_metadata.dt, this will serve as our
    # source of the "original" data dt information.


def subsample_scene_dt(scene: Scene, desired_dt: float) -> None:
    dt_ratio: float = desired_dt / scene.dt
    if not dt_ratio.is_integer():
        raise ValueError(
            f"Cannot subsample scene: {desired_dt} is not integer divisible by {scene.dt} for {str(scene)}"
        )

    dt_factor: int = int(dt_ratio)

    # E.g., the scene is currently at dt = 0.1s (10 Hz),
    # but we want desired_dt = 0.5s (2 Hz).
    scene.length_timesteps = (scene.length_timesteps - 1) // dt_factor + 1
    agent_presence: List[List[AgentMetadata]] = [
        [] for _ in range(scene.length_timesteps)
    ]
    for agent in scene.agents:
        # Need to be careful with the first timestep, since agents can have
        # first timesteps that are not exactly divisible by the dt_factor.
        agent.first_timestep = ceil(agent.first_timestep / dt_factor)
        agent.last_timestep //= dt_factor

        for scene_ts in range(agent.first_timestep, agent.last_timestep + 1):
            agent_presence[scene_ts].append(agent)

    scene.update_agent_info(scene.agents, agent_presence)
    scene.dt = desired_dt
    # Note we do not touch scene_info.env_metadata.dt, this will serve as our
    # source of the "original" data dt information.



import numpy as np


class MapPatch:
    def __init__(
        self,
        data: np.ndarray,
        rot_angle: float,
        crop_size: int,
        resolution: float,
        raster_from_world_tf: np.ndarray,
    ) -> None:
        self.data = data
        self.rot_angle = rot_angle
        self.crop_size = crop_size
        self.resolution = resolution
        self.raster_from_world_tf = raster_from_world_tf



import numpy as np
import pandas as pd

from trajdata.augmentation.augmentation import DatasetAugmentation


class LowSpeedYawCorrection(DatasetAugmentation):
    def __init__(self, speed_threshold: float = 0.0) -> None:
        self.speed_threshold = speed_threshold

    def apply(self, scene_data_df: pd.DataFrame) -> None:
        speed = np.linalg.norm(scene_data_df[["vx", "vy"]], axis=1)

        scene_data_df["yaw_diffs"] = scene_data_df["heading"].diff()
        # Doing this because the first row is always nan
        scene_data_df["yaw_diffs"].iat[0] = 0.0

        agent_ids: pd.Series = scene_data_df.index.get_level_values(0).astype(
            "category"
        )

        # The point of the border mask is to catch data like this:
        # index    agent_id     vx    vy
        #     0           1    7.3   9.1
        #     1           2    0.0   0.0
        #                  ...
        # As implemented, we would currently only
        # return index 0 (since we chop off the top with the 1: in the slice below), but
        # we want to return 1 so that's why the + 1 at the end.
        border_mask: np.ndarray = np.concatenate(
            [[0], np.nonzero(agent_ids[1:] != agent_ids[:-1])[0] + 1]
        )

        scene_data_df["yaw_diffs"].iloc[border_mask] = 0.0
        scene_data_df["yaw_diffs"].iloc[speed < self.speed_threshold] = 0.0

        mask_arr = np.ones((len(scene_data_df),), dtype=np.bool)
        mask_arr[border_mask] = False
        scene_data_df["heading"].iloc[mask_arr] = 0.0

        scene_data_df.loc[:, "yaw_diffs"] += scene_data_df["heading"]
        scene_data_df.loc[:, "heading"] = scene_data_df.groupby("agent_id")[
            "yaw_diffs"
        ].cumsum()

        scene_data_df.drop(columns="yaw_diffs", inplace=True)



import pickle
from math import ceil, floor
from pathlib import Path
from typing import Callable, Dict, Final, List, Optional, Tuple

import dill
import kornia
import numpy as np
import pandas as pd
import torch
import zarr

from trajdata.augmentation.augmentation import Augmentation, DatasetAugmentation
from trajdata.caching.scene_cache import SceneCache
from trajdata.data_structures.agent import AgentMetadata, FixedExtent
from trajdata.data_structures.scene_metadata import Scene
from trajdata.maps import RasterizedMap, RasterizedMapMetadata
from trajdata.proto.vectorized_map_pb2 import VectorizedMap
from trajdata.utils import arr_utils

STATE_COLS: Final[List[str]] = ["x", "y", "vx", "vy", "ax", "ay"]
EXTENT_COLS: Final[List[str]] = ["length", "width", "height"]


class DataFrameCache(SceneCache):
    def __init__(
        self,
        cache_path: Path,
        scene: Scene,
        scene_ts: Optional[int] = 0,
        augmentations: Optional[List[Augmentation]] = None,
    ) -> None:
        """
        Data cache primarily based on pandas DataFrames,
        with Feather for fast agent data serialization
        and pickle for miscellaneous supporting objects.
        Maps are pre-rasterized and stored as Zarr arrays.
        """
        super().__init__(cache_path, scene, scene_ts, augmentations)

        agent_data_path: Path = self.scene_dir / DataFrameCache._agent_data_file(
            scene.dt
        )
        if not agent_data_path.exists():
            # Load the original dt agent data and then
            # interpolate it to the desired dt.
            self._load_agent_data(scene.env_metadata.dt)
            self.interpolate_data(scene.dt)
        else:
            # Load the data with the desired dt.
            self._load_agent_data(scene.dt)

        # Setting default data transformation parameters.
        self.reset_transforms()

        if augmentations:
            dataset_augments: List[DatasetAugmentation] = [
                augment
                for augment in augmentations
                if isinstance(augment, DatasetAugmentation)
            ]
            for aug in dataset_augments:
                aug.apply(self.scene_data_df)

    @staticmethod
    def _agent_data_file(scene_dt: float) -> str:
        return f"agent_data_dt{scene_dt:.2f}.feather"

    @staticmethod
    def _agent_data_index_file(scene_dt: float) -> str:
        return f"scene_index_dt{scene_dt:.2f}.pkl"

    # AGENT STATE DATA
    def _get_and_reorder_col_idxs(self) -> None:
        avail_extent_cols: List[str] = [
            col for col in self.scene_data_df.columns if col in EXTENT_COLS
        ]
        avail_heading_cols: List[str] = (
            ["heading"]
            if "heading" in self.scene_data_df.columns
            else ["sin_heading", "cos_heading"]
        )
        self.scene_data_df = self.scene_data_df[
            STATE_COLS + avail_heading_cols + avail_extent_cols
        ]

        self.column_dict: Dict[str, int] = {
            val: idx for idx, val in enumerate(self.scene_data_df.columns)
        }

        self.pos_cols = [self.column_dict["x"], self.column_dict["y"]]
        self.vel_cols = [self.column_dict["vx"], self.column_dict["vy"]]
        self.acc_cols = [self.column_dict["ax"], self.column_dict["ay"]]
        if "sin_heading" in self.column_dict:
            self.heading_cols = [
                self.column_dict["sin_heading"],
                self.column_dict["cos_heading"],
            ]
        else:
            self.heading_cols = [self.column_dict["heading"]]

        self.state_cols = (
            self.pos_cols + self.vel_cols + self.acc_cols + self.heading_cols
        )
        self._state_dim = len(self.state_cols)

        # While it may seem that obs_dim == _state_dim, obs_dim is meant to
        # track the output dimension of states (which could differ via
        # returing sin and cos of heading rather than just heading itself).
        self.obs_dim = self._state_dim

        self.extent_cols: List[int] = list()
        for extent_name in ["length", "width", "height"]:
            if extent_name in self.column_dict:
                self.extent_cols.append(self.column_dict[extent_name])

    def _load_agent_data(self, scene_dt: float) -> None:
        self.scene_data_df: pd.DataFrame = pd.read_feather(
            self.scene_dir / DataFrameCache._agent_data_file(scene_dt),
            use_threads=False,
        ).set_index(["agent_id", "scene_ts"])

        with open(
            self.scene_dir / DataFrameCache._agent_data_index_file(scene_dt), "rb"
        ) as f:
            self.index_dict: Dict[Tuple[str, int], int] = pickle.load(f)

        self._get_and_reorder_col_idxs()

    def write_cache_to_disk(self) -> None:
        with open(
            self.scene_dir / DataFrameCache._agent_data_index_file(self.dt), "wb"
        ) as f:
            pickle.dump(self.index_dict, f)

        self.scene_data_df.reset_index().to_feather(
            self.scene_dir / DataFrameCache._agent_data_file(self.dt)
        )

    @staticmethod
    def save_agent_data(
        agent_data: pd.DataFrame,
        cache_path: Path,
        scene: Scene,
    ) -> None:
        scene_cache_dir: Path = cache_path / scene.env_name / scene.name
        scene_cache_dir.mkdir(parents=True, exist_ok=True)

        index_dict: Dict[Tuple[str, int], int] = {
            val: idx for idx, val in enumerate(agent_data.index)
        }
        with open(
            scene_cache_dir / DataFrameCache._agent_data_index_file(scene.dt), "wb"
        ) as f:
            pickle.dump(index_dict, f)

        agent_data.reset_index().to_feather(
            scene_cache_dir / DataFrameCache._agent_data_file(scene.dt)
        )

    def get_value(self, agent_id: str, scene_ts: int, attribute: str) -> float:
        col_idx: int = self.column_dict[attribute]
        value: float = self.scene_data_df.iat[
            self.index_dict[(agent_id, scene_ts)], col_idx
        ]

        if "y" in attribute:
            x_value: float = self.scene_data_df.iat[
                self.index_dict[(agent_id, scene_ts)], col_idx - 1
            ]

            transformed_pair: np.ndarray = self._transform_pair(
                np.array([[x_value, value]]), (col_idx - 1, col_idx)
            )
            return transformed_pair[0, 1].item()
        elif "x" in attribute:
            y_value: float = self.scene_data_df.iat[
                self.index_dict[(agent_id, scene_ts)], col_idx + 1
            ]

            transformed_pair: np.ndarray = self._transform_pair(
                np.array([[value, y_value]]), (col_idx, col_idx + 1)
            )
            return transformed_pair[0, 0].item()
        else:
            # The 0.0 and 0 here are just placeholders.
            transformed_pair: np.ndarray = self._transform_pair(
                np.array([[value, 0.0]]), (col_idx, 0)
            )
            return transformed_pair[0, 0].item()

    def get_state(self, agent_id: str, scene_ts: int) -> np.ndarray:
        state = self.scene_data_df.iloc[
            self.index_dict[(agent_id, scene_ts)], : self._state_dim
        ].to_numpy()

        return self._transform_state(state)

    def get_states(self, agent_ids: List[str], scene_ts: int) -> np.ndarray:
        row_idxs: List[int] = [
            self.index_dict[(agent_id, scene_ts)] for agent_id in agent_ids
        ]

        states = self.scene_data_df.iloc[row_idxs, : self._state_dim].to_numpy()

        return self._transform_state(states)

    def transform_data(self, **kwargs) -> None:
        if "shift_mean_to" in kwargs:
            # This standardizes the scene to be relative to the agent being predicted.
            self._transf_mean = kwargs["shift_mean_to"]

        if "rotate_by" in kwargs:
            # This rotates the scene so that the predicted agent's current
            # heading aligns with the x-axis.
            agent_heading: float = kwargs["rotate_by"]
            self._transf_rotmat: np.ndarray = np.array(
                [
                    [np.cos(agent_heading), -np.sin(agent_heading)],
                    [np.sin(agent_heading), np.cos(agent_heading)],
                ]
            )

        if "sincos_heading" in kwargs:
            self._sincos_heading = True
            self.obs_dim += 1

    def reset_transforms(self) -> None:
        self._transf_mean: Optional[np.ndarray] = None
        self._transf_rotmat: Optional[np.ndarray] = None
        self._sincos_heading: bool = False

        self.obs_dim = self._state_dim

    def _transform_state(self, data: np.ndarray) -> np.ndarray:
        state: np.ndarray = data.copy()  # Don't want to alter the original data.

        if len(state.shape) == 1:
            state = state[np.newaxis, :]

        if self._transf_mean is not None:
            # Shift to zero mean, but leave heading
            # standardization for if rotation is also requested (below).
            state[..., :-1] -= self._transf_mean[:-1]

        if self._transf_rotmat is not None:
            state[..., self.pos_cols] = state[..., self.pos_cols] @ self._transf_rotmat
            state[..., self.vel_cols] = state[..., self.vel_cols] @ self._transf_rotmat
            state[..., self.acc_cols] = state[..., self.acc_cols] @ self._transf_rotmat
            state[..., -1] -= self._transf_mean[-1]

        if self._sincos_heading:
            sin_heading: np.ndarray = np.sin(state[..., [-1]])
            cos_heading: np.ndarray = np.cos(state[..., [-1]])
            state = np.concatenate([state[..., :-1], sin_heading, cos_heading], axis=-1)

        return state[0] if len(data.shape) == 1 else state

    def _transform_pair(
        self, data: np.ndarray, col_idxs: Tuple[int, int]
    ) -> np.ndarray:
        state: np.ndarray = data.copy()  # Don't want to alter the original data.

        if len(state.shape) == 1:
            state = state[np.newaxis, :]

        if self._transf_mean is not None:
            state -= self._transf_mean[col_idxs]

        if (
            self._transf_rotmat is not None
            and col_idxs[0] in self.pos_cols + self.vel_cols + self.acc_cols
        ):
            state = state @ self._transf_rotmat

        return state[0] if len(data.shape) == 1 else state

    def _upsample_data(
        self, new_index: pd.MultiIndex, upsample_dt_ratio: float, method: str
    ) -> pd.DataFrame:
        upsample_dt_factor: int = int(upsample_dt_ratio)

        interpolated_df: pd.DataFrame = pd.DataFrame(
            index=new_index, columns=self.scene_data_df.columns
        )
        interpolated_df = interpolated_df.astype(self.scene_data_df.dtypes.to_dict())

        scene_data: np.ndarray = self.scene_data_df.to_numpy()
        unwrapped_heading: np.ndarray = np.unwrap(self.scene_data_df["heading"])

        # Getting the data initially in the new df, making sure to unwrap angles above
        # in preparation for interpolation.
        scene_data_idxs: np.ndarray = np.nonzero(
            new_index.get_level_values("scene_ts") % upsample_dt_factor == 0
        )[0]
        interpolated_df.iloc[scene_data_idxs] = scene_data
        interpolated_df.iloc[
            scene_data_idxs, self.column_dict["heading"]
        ] = unwrapped_heading

        # Interpolation.
        interpolated_df.interpolate(method=method, limit_area="inside", inplace=True)

        # Wrapping angles back to [-pi, pi).
        interpolated_df.iloc[:, self.column_dict["heading"]] = arr_utils.angle_wrap(
            interpolated_df.iloc[:, self.column_dict["heading"]]
        )

        return interpolated_df

    def _downsample_data(
        self, new_index: pd.MultiIndex, downsample_dt_ratio: float
    ) -> pd.DataFrame:
        downsample_dt_factor: int = int(downsample_dt_ratio)

        subsample_index: pd.MultiIndex = new_index.set_levels(
            new_index.levels[1] * downsample_dt_factor, level=1
        )

        subsampled_df: pd.DataFrame = self.scene_data_df.reindex(
            index=subsample_index
        ).set_index(new_index)

        return subsampled_df

    def interpolate_data(self, desired_dt: float, method: str = "linear") -> None:
        upsample_dt_ratio: float = self.scene.env_metadata.dt / desired_dt
        downsample_dt_ratio: float = desired_dt / self.scene.env_metadata.dt
        if not upsample_dt_ratio.is_integer() and not downsample_dt_ratio.is_integer():
            raise ValueError(
                f"{str(self.scene)}'s dt of {self.scene.dt}s "
                f"is not integer divisible by the desired dt {desired_dt}s."
            )

        agent_info_dict: Dict[str, AgentMetadata] = {
            agent.name: agent for agent in self.scene.agents
        }
        new_index: pd.MultiIndex = pd.MultiIndex.from_tuples(
            [
                (agent_name, scene_ts)
                for agent_name in self.scene_data_df.index.unique(level=0)
                for scene_ts in range(
                    # These have already been multiplied by dt_factor earlier.
                    agent_info_dict[agent_name].first_timestep,
                    agent_info_dict[agent_name].last_timestep + 1,
                )
            ],
            names=["agent_id", "scene_ts"],
        )

        if upsample_dt_ratio >= 1:
            interpolated_df = self._upsample_data(new_index, upsample_dt_ratio, method)
        elif downsample_dt_ratio >= 1:
            interpolated_df = self._downsample_data(new_index, downsample_dt_ratio)

        self.dt = desired_dt
        self.scene_data_df = interpolated_df
        self.index_dict: Dict[Tuple[str, int], int] = {
            val: idx for idx, val in enumerate(self.scene_data_df.index)
        }

    def get_agent_history(
        self,
        agent_info: AgentMetadata,
        scene_ts: int,
        history_sec: Tuple[Optional[float], Optional[float]],
    ) -> Tuple[np.ndarray, np.ndarray]:
        # We don't have to check the mins here because our data_index filtering in dataset.py already
        # took care of it.
        first_index_incl: int
        last_index_incl: int = self.index_dict[(agent_info.name, scene_ts)]
        if history_sec[1] is not None:
            max_history: int = floor(history_sec[1] / self.dt)
            first_index_incl = self.index_dict[
                (
                    agent_info.name,
                    max(scene_ts - max_history, agent_info.first_timestep),
                )
            ]
        else:
            first_index_incl = self.index_dict[
                (agent_info.name, agent_info.first_timestep)
            ]

        agent_history_df: pd.DataFrame = self.scene_data_df.iloc[
            first_index_incl : last_index_incl + 1
        ]

        agent_extent_np: np.ndarray
        if isinstance(agent_info.extent, FixedExtent):
            agent_extent_np = agent_info.extent.get_extents(
                self.scene_ts - agent_history_df.shape[0] + 1, self.scene_ts
            )
        else:
            agent_extent_np = agent_history_df.iloc[:, self.extent_cols].to_numpy()

        return (
            self._transform_state(
                agent_history_df.iloc[:, : self._state_dim].to_numpy()
            ),
            agent_extent_np,
        )

    def get_agent_future(
        self,
        agent_info: AgentMetadata,
        scene_ts: int,
        future_sec: Tuple[Optional[float], Optional[float]],
    ) -> Tuple[np.ndarray, np.ndarray]:
        # We don't have to check the mins here because our data_index filtering in dataset.py already
        # took care of it.
        if scene_ts >= agent_info.last_timestep:
            # Extent shape = 3
            return np.zeros((0, self.obs_dim)), np.zeros((0, 3))

        first_index_incl: int = self.index_dict[(agent_info.name, scene_ts + 1)]
        last_index_incl: int
        if future_sec[1] is not None:
            max_future = floor(future_sec[1] / self.dt)
            last_index_incl = self.index_dict[
                (agent_info.name, min(scene_ts + max_future, agent_info.last_timestep))
            ]
        else:
            last_index_incl = self.index_dict[
                (agent_info.name, agent_info.last_timestep)
            ]

        agent_future_df: pd.DataFrame = self.scene_data_df.iloc[
            first_index_incl : last_index_incl + 1
        ]

        agent_extent_np: np.ndarray
        if isinstance(agent_info.extent, FixedExtent):
            agent_extent_np: np.ndarray = agent_info.extent.get_extents(
                self.scene_ts + 1, self.scene_ts + agent_future_df.shape[0]
            )
        else:
            agent_extent_np = agent_future_df.iloc[:, self.extent_cols].to_numpy()

        return (
            self._transform_state(
                agent_future_df.iloc[:, : self._state_dim].to_numpy()
            ),
            agent_extent_np,
        )

    def get_agents_history(
        self,
        scene_ts: int,
        agents: List[AgentMetadata],
        history_sec: Tuple[Optional[float], Optional[float]],
    ) -> Tuple[List[np.ndarray], List[np.ndarray], np.ndarray]:
        first_timesteps = np.array(
            [agent.first_timestep for agent in agents], dtype=np.long
        )
        if history_sec[1] is not None:
            max_history: int = floor(history_sec[1] / self.dt)
            first_timesteps = np.maximum(scene_ts - max_history, first_timesteps)

        first_index_incl: np.ndarray = np.array(
            [
                self.index_dict[(agent.name, first_timesteps[idx])]
                for idx, agent in enumerate(agents)
            ],
            dtype=np.long,
        )
        last_index_incl: np.ndarray = np.array(
            [self.index_dict[(agent.name, scene_ts)] for agent in agents], dtype=np.long
        )

        concat_idxs = arr_utils.vrange(first_index_incl, last_index_incl + 1)
        neighbor_data_df: pd.DataFrame = self.scene_data_df.iloc[concat_idxs, :]

        neighbor_history_lens_np = last_index_incl - first_index_incl + 1

        neighbor_histories_np = self._transform_state(
            neighbor_data_df.iloc[:, : self._state_dim].to_numpy()
        )
        # The last one will always be empty because of what cumsum returns.
        neighbor_histories: List[np.ndarray] = np.vsplit(
            neighbor_histories_np, neighbor_history_lens_np.cumsum()
        )[:-1]

        neighbor_extents: List[np.ndarray]
        if self.extent_cols:
            neighbor_extents_np = neighbor_data_df.iloc[:, self.extent_cols].to_numpy()
            # The last one will always be empty because of what cumsum returns.
            neighbor_extents = np.vsplit(
                neighbor_extents_np, neighbor_history_lens_np.cumsum()
            )[:-1]
        else:
            neighbor_extents = [
                agent.extent.get_extents(
                    self.scene_ts - neighbor_history_lens_np[idx].item() + 1,
                    self.scene_ts,
                )
                for idx, agent in enumerate(agents)
            ]

        return (
            neighbor_histories,
            neighbor_extents,
            neighbor_history_lens_np,
        )

    def get_agents_future(
        self,
        scene_ts: int,
        agents: List[AgentMetadata],
        future_sec: Tuple[Optional[float], Optional[float]],
    ) -> Tuple[List[np.ndarray], List[np.ndarray], np.ndarray]:
        last_timesteps = np.array(
            [agent.last_timestep for agent in agents], dtype=np.long
        )

        first_timesteps = np.minimum(scene_ts + 1, last_timesteps)

        if future_sec[1] is not None:
            max_future: int = floor(future_sec[1] / self.dt)
            last_timesteps = np.minimum(scene_ts + max_future, last_timesteps)

        first_index_incl: np.ndarray = np.array(
            [
                self.index_dict[(agent.name, first_timesteps[idx])]
                for idx, agent in enumerate(agents)
            ],
            dtype=np.long,
        )
        last_index_incl: np.ndarray = np.array(
            [
                self.index_dict[(agent.name, last_timesteps[idx])]
                for idx, agent in enumerate(agents)
            ],
            dtype=np.long,
        )

        concat_idxs = arr_utils.vrange(first_index_incl, last_index_incl + 1)
        neighbor_data_df: pd.DataFrame = self.scene_data_df.iloc[concat_idxs, :]

        neighbor_future_lens_np = last_index_incl - first_index_incl + 1

        neighbor_futures_np = self._transform_state(
            neighbor_data_df.iloc[:, : self._state_dim].to_numpy()
        )
        # The last one will always be empty because of what cumsum returns.
        neighbor_futures: List[np.ndarray] = np.vsplit(
            neighbor_futures_np, neighbor_future_lens_np.cumsum()
        )[:-1]

        neighbor_extents: List[np.ndarray]
        if self.extent_cols:
            neighbor_extents_np = neighbor_data_df.iloc[:, self.extent_cols].to_numpy()
            # The last one will always be empty because of what cumsum returns.
            neighbor_extents = np.vsplit(
                neighbor_extents_np, neighbor_future_lens_np.cumsum()
            )[:-1]
        else:
            neighbor_extents = [
                agent.extent.get_extents(
                    self.scene_ts - neighbor_future_lens_np[idx].item() + 1,
                    self.scene_ts,
                )
                for idx, agent in enumerate(agents)
            ]

        return (
            neighbor_futures,
            neighbor_extents,
            neighbor_future_lens_np,
        )

    # MAPS
    @staticmethod
    def get_maps_path(cache_path: Path, env_name: str) -> Path:
        return cache_path / env_name / "maps"

    @staticmethod
    def are_maps_cached(cache_path: Path, env_name: str) -> bool:
        return DataFrameCache.get_maps_path(cache_path, env_name).exists()

    @staticmethod
    def get_map_paths(
        cache_path: Path, env_name: str, map_name: str, resolution: float
    ) -> Tuple[Path, Path, Path, Path]:
        maps_path: Path = DataFrameCache.get_maps_path(cache_path, env_name)

        vector_map_path: Path = maps_path / f"{map_name}.pb"
        raster_map_path: Path = maps_path / f"{map_name}_{resolution:.2f}px_m.zarr"
        raster_metadata_path: Path = maps_path / f"{map_name}_{resolution:.2f}px_m.dill"

        return maps_path, vector_map_path, raster_map_path, raster_metadata_path

    @staticmethod
    def is_map_cached(
        cache_path: Path, env_name: str, map_name: str, resolution: float
    ) -> bool:
        (
            maps_path,
            vector_map_path,
            raster_map_path,
            raster_metadata_path,
        ) = DataFrameCache.get_map_paths(cache_path, env_name, map_name, resolution)
        return (
            maps_path.exists()
            and vector_map_path.exists()
            and raster_metadata_path.exists()
            and raster_map_path.exists()
        )

    @staticmethod
    def cache_map(
        cache_path: Path, vec_map: VectorizedMap, map_obj: RasterizedMap, env_name: str
    ) -> None:
        (
            maps_path,
            vector_map_path,
            raster_map_path,
            raster_metadata_path,
        ) = DataFrameCache.get_map_paths(
            cache_path, env_name, map_obj.metadata.name, map_obj.metadata.resolution
        )

        # Ensuring the maps directory exists.
        maps_path.mkdir(parents=True, exist_ok=True)

        # Saving the vectorized map data.
        with open(vector_map_path, "wb") as f:
            f.write(vec_map.SerializeToString())

        # Saving the rasterized map data.
        zarr.save(raster_map_path, map_obj.data)

        # Saving the rasterized map metadata.
        with open(raster_metadata_path, "wb") as f:
            dill.dump(map_obj.metadata, f)

    @staticmethod
    def cache_map_layers(
        cache_path: Path,
        vec_map: VectorizedMap,
        map_info: RasterizedMapMetadata,
        layer_fn: Callable[[str], np.ndarray],
        env_name: str,
    ) -> None:
        (
            maps_path,
            vector_map_path,
            raster_map_path,
            raster_metadata_path,
        ) = DataFrameCache.get_map_paths(
            cache_path, env_name, map_info.name, map_info.resolution
        )

        # Ensuring the maps directory exists.
        maps_path.mkdir(parents=True, exist_ok=True)

        # Saving the vectorized map data.
        with open(vector_map_path, "wb") as f:
            f.write(vec_map.SerializeToString())

        # Saving the rasterized map data.
        disk_data = zarr.open_array(raster_map_path, mode="w", shape=map_info.shape)
        for idx, layer_name in enumerate(map_info.layers):
            disk_data[idx] = layer_fn(layer_name)

        # Saving the rasterized map metadata.
        with open(raster_metadata_path, "wb") as f:
            dill.dump(map_info, f)

    def pad_map_patch(
        self,
        patch: np.ndarray,
        #                 top, bot, left, right
        patch_sides: Tuple[int, int, int, int],
        patch_size: int,
        map_dims: Tuple[int, int, int],
    ) -> np.ndarray:
        if patch.shape[-2:] == (patch_size, patch_size):
            return patch

        top, bot, left, right = patch_sides
        channels, height, width = map_dims

        # If we're off the map, just return zeros in the
        # desired size of the patch.
        if bot <= 0 or top >= height or right <= 0 or left >= width:
            return np.zeros((channels, patch_size, patch_size))

        pad_top, pad_bot, pad_left, pad_right = 0, 0, 0, 0
        if top < 0:
            pad_top = 0 - top
        if bot >= height:
            pad_bot = bot - height
        if left < 0:
            pad_left = 0 - left
        if right >= width:
            pad_right = right - width

        return np.pad(patch, [(0, 0), (pad_top, pad_bot), (pad_left, pad_right)])

    def load_map_patch(
        self,
        world_x: float,
        world_y: float,
        desired_patch_size: int,
        resolution: float,
        offset_xy: Tuple[float, float],
        agent_heading: float,
        return_rgb: bool,
        rot_pad_factor: float = 1.0,
        no_map_val: float = 0.0,
    ) -> Tuple[np.ndarray, np.ndarray, bool]:
        (
            maps_path,
            _,
            raster_map_path,
            raster_metadata_path,
        ) = DataFrameCache.get_map_paths(
            self.path, self.scene.env_name, self.scene.location, resolution
        )
        if not maps_path.exists():
            # This dataset (or location) does not have any maps,
            # so we return an empty map.
            patch_size: int = ceil((rot_pad_factor * desired_patch_size) / 2) * 2

            return (
                np.full(
                    (1 if not return_rgb else 3, patch_size, patch_size),
                    fill_value=no_map_val,
                ),
                np.eye(3),
                False,
            )

        with open(raster_metadata_path, "rb") as f:
            map_info: RasterizedMapMetadata = dill.load(f)

        raster_from_world_tf: np.ndarray = map_info.map_from_world
        map_coords: np.ndarray = map_info.map_from_world @ np.array(
            [world_x, world_y, 1.0]
        )
        map_x, map_y = map_coords[0].item(), map_coords[1].item()

        raster_from_world_tf = (
            np.array(
                [
                    [1.0, 0.0, -map_x],
                    [0.0, 1.0, -map_y],
                    [0.0, 0.0, 1.0],
                ]
            )
            @ raster_from_world_tf
        )

        # This first size is how much of the map we
        # need to extract to match the requested metric size (meters x meters) of
        # the patch.
        data_patch_size: int = ceil(
            desired_patch_size * map_info.resolution / resolution
        )

        # Incorporating offsets.
        if offset_xy != (0.0, 0.0):
            # x is negative here because I am moving the map
            # center so that the agent ends up where the user wishes
            # (the agent is pinned from the end user's perspective).
            map_offset: Tuple[float, float] = (
                -offset_xy[0] * data_patch_size // 2,
                offset_xy[1] * data_patch_size // 2,
            )

            rotated_offset: np.ndarray = (
                arr_utils.rotation_matrix(agent_heading) @ map_offset
            )

            off_x = rotated_offset[0]
            off_y = rotated_offset[1]

            map_x += off_x
            map_y += off_y

            raster_from_world_tf = (
                np.array(
                    [
                        [1.0, 0.0, -off_x],
                        [0.0, 1.0, -off_y],
                        [0.0, 0.0, 1.0],
                    ]
                )
                @ raster_from_world_tf
            )

        # This is the size of the patch taking into account expansion to allow for
        # rotation to match the agent's heading. We also ensure the final size is
        # divisible by two so that the // 2 below does not chop any information off.
        data_with_rot_pad_size: int = ceil((rot_pad_factor * data_patch_size) / 2) * 2

        disk_data = zarr.open_array(raster_map_path, mode="r")

        map_x = round(map_x)
        map_y = round(map_y)

        # Half of the patch's side length.
        half_extent: int = data_with_rot_pad_size // 2

        top: int = map_y - half_extent
        bot: int = map_y + half_extent
        left: int = map_x - half_extent
        right: int = map_x + half_extent

        data_patch: np.ndarray = self.pad_map_patch(
            disk_data[
                ...,
                max(top, 0) : min(bot, disk_data.shape[1]),
                max(left, 0) : min(right, disk_data.shape[2]),
            ],
            (top, bot, left, right),
            data_with_rot_pad_size,
            disk_data.shape,
        )

        if return_rgb:
            rgb_groups = map_info.layer_rgb_groups
            data_patch = np.stack(
                [
                    np.amax(data_patch[rgb_groups[0]], axis=0),
                    np.amax(data_patch[rgb_groups[1]], axis=0),
                    np.amax(data_patch[rgb_groups[2]], axis=0),
                ],
            )

        if desired_patch_size != data_patch_size:
            scale_factor: float = desired_patch_size / data_patch_size
            data_patch = (
                kornia.geometry.rescale(
                    torch.from_numpy(data_patch).unsqueeze(0),
                    scale_factor,
                    # Default align_corners value, just putting it to remove warnings
                    align_corners=False,
                    antialias=True,
                )
                .squeeze(0)
                .numpy()
            )

            raster_from_world_tf = (
                np.array(
                    [
                        [1 / scale_factor, 0.0, 0.0],
                        [0.0, 1 / scale_factor, 0.0],
                        [0.0, 0.0, 1.0],
                    ]
                )
                @ raster_from_world_tf
            )

        return data_patch, raster_from_world_tf, True



from typing import Dict, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def plot_sim_stats(
    stats: Dict[str, Dict[str, Tuple[np.ndarray, np.ndarray]]],
    show: bool = True,
    close: bool = True,
):
    fig, axes = plt.subplots(nrows=len(stats), ncols=2, figsize=(4, 8))

    axes[0, 0].set_title("Ground Truth")
    axes[0, 1].set_title("Simulated")

    for row, scene_stats in enumerate(stats.values()):
        histogram, bins = scene_stats["gt"]
        axes[row, 0].hist(histogram, bins, linewidth=0.5, edgecolor="white")

        histogram, bins = scene_stats["sim"]
        axes[row, 1].hist(histogram, bins, linewidth=0.5, edgecolor="white")

    plt.tight_layout()

    if show:
        plt.show()

    if close:
        plt.close(fig)



from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Set, Tuple, Type, Union

from trajdata.caching import EnvCache, SceneCache
from trajdata.data_structures import (
    AgentMetadata,
    EnvMetadata,
    Scene,
    SceneMetadata,
    SceneTag,
)


class RawDataset:
    def __init__(
        self, env_name: str, data_dir: str, parallelizable: bool, has_maps: bool
    ) -> None:
        metadata = self.compute_metadata(env_name, data_dir)

        self.metadata = metadata
        self.name = metadata.name
        self.scene_tags = metadata.scene_tags
        self.dataset_obj = None

        self.parallelizable = parallelizable
        self.has_maps = has_maps

    def compute_metadata(self, env_name: str, data_dir: str) -> EnvMetadata:
        raise NotImplementedError()

    def get_matching_scene_tags(self, query: Set[str]) -> List[SceneTag]:
        return [scene_tag for scene_tag in self.scene_tags if scene_tag.contains(query)]

    def load_dataset_obj(self, verbose: bool = False) -> None:
        raise NotImplementedError()

    def del_dataset_obj(self) -> None:
        del self.dataset_obj
        self.dataset_obj = None

    def _get_matching_scenes_from_cache(
        self,
        scene_tag: SceneTag,
        scene_desc_contains: Optional[List[str]],
        env_cache: EnvCache,
    ) -> List[Scene]:
        raise NotImplementedError()

    def _get_matching_scenes_from_obj(
        self,
        scene_tag: SceneTag,
        scene_desc_contains: Optional[List[str]],
        env_cache: EnvCache,
    ) -> List[SceneMetadata]:
        raise NotImplementedError()

    def cache_all_scenes_list(
        self, env_cache: EnvCache, all_scenes_list: List[NamedTuple]
    ) -> None:
        env_cache.save_env_scenes_list(self.name, all_scenes_list)

    def get_matching_scenes(
        self,
        scene_tag: SceneTag,
        scene_desc_contains: Optional[List[str]],
        env_cache: EnvCache,
        rebuild_cache: bool,
    ) -> Union[List[Scene], List[SceneMetadata]]:
        if self.dataset_obj is None and not rebuild_cache:
            return self._get_matching_scenes_from_cache(
                scene_tag, scene_desc_contains, env_cache
            )
        else:
            return self._get_matching_scenes_from_obj(
                scene_tag, scene_desc_contains, env_cache
            )

    def get_scene(self, scene_info: SceneMetadata) -> Scene:
        raise NotImplementedError()

    def get_agent_info(
        self, scene: Scene, cache_path: Path, cache_class: Type[SceneCache]
    ) -> Tuple[List[AgentMetadata], List[List[AgentMetadata]]]:
        raise NotImplementedError()

    def cache_maps(
        self,
        cache_path: Path,
        map_cache_class: Type[SceneCache],
        map_params: Dict[str, Any],
    ) -> None:
        """
        resolution is in pixels per meter.
        """
        raise NotImplementedError()



import pandas as pd

from trajdata.data_structures.batch import AgentBatch, SceneBatch


class Augmentation:
    def __init__(self) -> None:
        raise NotImplementedError()


class DatasetAugmentation(Augmentation):
    def apply(self, scene_data_df: pd.DataFrame) -> None:
        raise NotImplementedError()


class BatchAugmentation(Augmentation):
    def apply_agent(self, agent_batch: AgentBatch) -> None:
        raise NotImplementedError()

    def apply_scene(self, scene_batch: SceneBatch) -> None:
        raise NotImplementedError()



from math import ceil
from typing import List, Optional, Set, Tuple

from trajdata.data_structures.agent import AgentMetadata, AgentType


def agent_types(
    agents: List[AgentMetadata], no_types: Set[AgentType], only_types: Set[AgentType]
) -> List[AgentMetadata]:
    agents_list: List[AgentMetadata] = agents

    if no_types is not None:
        agents_list = [agent for agent in agents_list if agent.type not in no_types]

    if only_types is not None:
        agents_list = [agent for agent in agents_list if agent.type in only_types]

    return agents_list


def all_agents_excluded_types(
    no_types: Optional[List[AgentType]], agents: List[AgentMetadata]
) -> bool:
    return no_types is not None and all(
        agent_info.type in no_types for agent_info in agents
    )


def no_agent_included_types(
    only_types: Optional[List[AgentType]], agents: List[AgentMetadata]
) -> bool:
    return only_types is not None and all(
        agent_info.type not in only_types for agent_info in agents
    )


def get_valid_ts(
    agent_info: AgentMetadata,
    dt: float,
    history_sec: Tuple[Optional[float], Optional[float]],
    future_sec: Tuple[Optional[float], Optional[float]],
) -> Tuple[int, int]:
    """The returned timesteps are both inclusive.

    Args:
        agent_info (AgentMetadata): _description_
        dt (float): _description_
        history_sec (Tuple[Optional[float], Optional[float]]): _description_
        future_sec (Tuple[Optional[float], Optional[float]]): _description_

    Returns:
        Tuple[int, int]: _description_
    """
    first_valid_ts = agent_info.first_timestep
    if history_sec[0] is not None:
        min_history = ceil(history_sec[0] / dt)
        first_valid_ts += min_history

    last_valid_ts = agent_info.last_timestep
    if future_sec[0] is not None:
        min_future = ceil(future_sec[0] / dt)
        last_valid_ts -= min_future

    return first_valid_ts, last_valid_ts


def satisfies_history(
    agent_info: AgentMetadata,
    ts: int,
    dt: float,
    history_sec: Tuple[Optional[float], Optional[float]],
) -> bool:
    if history_sec[0] is not None:
        min_history = ceil(history_sec[0] / dt)
        agent_history_satisfies = ts - agent_info.first_timestep >= min_history
    else:
        agent_history_satisfies = True

    return agent_history_satisfies


def satisfies_future(
    agent_info: AgentMetadata,
    ts: int,
    dt: float,
    future_sec: Tuple[Optional[float], Optional[float]],
) -> bool:
    if future_sec[0] is not None:
        min_future = ceil(future_sec[0] / dt)
        agent_future_satisfies = agent_info.last_timestep - ts >= min_future
    else:
        agent_future_satisfies = True

    return agent_future_satisfies


def satisfies_times(
    agent_info: AgentMetadata,
    ts: int,
    dt: float,
    history_sec: Tuple[Optional[float], Optional[float]],
    future_sec: Tuple[Optional[float], Optional[float]],
) -> bool:
    agent_history_satisfies = satisfies_history(agent_info, ts, dt, history_sec)
    agent_future_satisfies = satisfies_future(agent_info, ts, dt, future_sec)
    return agent_history_satisfies and agent_future_satisfies


def no_agent_satisfies_time(
    ts: int,
    dt: float,
    history_sec: Tuple[Optional[float], Optional[float]],
    future_sec: Tuple[Optional[float], Optional[float]],
    agents: List[AgentMetadata],
) -> bool:
    return all(
        not satisfies_times(agent_info, ts, dt, history_sec, future_sec)
        for agent_info in agents
    )
